''' Utilities for dealing with sam files. '''

import re
import subprocess
import os
import shutil
import heapq
import contextlib
import copy
import functools

from collections import Counter
from pathlib import Path

import pysam
import numpy as np

from . import utilities
from . import fastq
from . import fasta
from . import mapping_tools
from . import interval

BAM_CMATCH = 0     # M
BAM_CINS = 1       # I
BAM_CDEL = 2       # D
BAM_CREF_SKIP = 3  # N
BAM_CSOFT_CLIP = 4 # S
BAM_CHARD_CLIP = 5 # H
BAM_CPAD = 6       # P
BAM_CEQUAL = 7     # =
BAM_CDIFF = 8      # X

op_to_char = {
    BAM_CMATCH:     'M',
    BAM_CINS:       'I',
    BAM_CDEL:       'D',
    BAM_CREF_SKIP:  'N',
    BAM_CSOFT_CLIP: 'S',
    BAM_CHARD_CLIP: 'H',
    BAM_CPAD:       'P',
    BAM_CEQUAL:     '=',
    BAM_CDIFF:      'X',
}
# Want to be able to lookup with int or char keys, so make every relevant char
# return itself.
for v in list(op_to_char.values()):
    op_to_char[v] = v

read_consuming_ops = {
    BAM_CMATCH,
    BAM_CINS,
    BAM_CSOFT_CLIP,
    BAM_CEQUAL,
    BAM_CDIFF,
}

ref_consuming_ops = {
    BAM_CMATCH,
    BAM_CDEL,
    BAM_CEQUAL,
    BAM_CDIFF,
    BAM_CREF_SKIP,
}

def get_strand(mapping):
    if mapping.is_reverse:
        strand = '-'
    else:
        strand = '+'
    return strand

def get_opposite_strand(mapping):
    if mapping.is_reverse:
        strand = '+'
    else:
        strand = '-'
    return strand

def get_original_seq(mapping):
    if mapping.is_reverse:
        original_seq = utilities.reverse_complement(mapping.query_sequence)
    else:
        original_seq = mapping.query_sequence
    return original_seq

def get_original_qual(mapping):
    if mapping.is_reverse:
        original_qual = mapping.query_qualities[::-1]
    else:
        original_qual = mapping.query_qualities
    return original_qual

def unmapped_aligned_read(qname):
    aligned_read = pysam.AlignedRead()
    aligned_read.qname = qname
    aligned_read.flag = 0x4
    aligned_read.rname = -1
    aligned_read.pos = -1
    aligned_read.mapq = 0
    aligned_read.cigar = None
    aligned_read.rnext = -1
    aligned_read.pnext = -1
    aligned_read.tlen = 0
    aligned_read.seq = '*'
    aligned_read.qual = '*'
    return aligned_read

def splice_in_name(line, new_name):
    return '\t'.join([new_name] + line.split('\t')[1:])

cigar_block = re.compile(r'(\d+)([MIDNSHP=X])')

def cigar_string_to_blocks(cigar_string):
    """ Decomposes a CIGAR string into a list of its operations. """
    return [(int(l), k) for l, k in cigar_block.findall(cigar_string)]

def total_reference_nucs(cigar):
    return sum(length for op, length in cigar if op in ref_consuming_ops)

def total_reference_nucs_except_splicing(cigar):
    return sum(length for op, length in cigar if op in ref_consuming_ops and op != BAM_CREF_SKIP)

def total_read_nucs(cigar):
    return sum(length for op, length in cigar if op in read_consuming_ops)

def contains_indel(read):
    kinds = [k for k, l in read.cigar]
    return (BAM_CINS in kinds or BAM_CDEL in kinds)

def indel_distance_from_edge(cigar):
    indel_indices = [i for i, (k, l) in enumerate(cigar) if k == BAM_CINS or k == BAM_CDEL]
    first_indel_index = min(indel_indices)
    ref_nucs_before = total_reference_nucs(cigar[:first_indel_index])
    last_indel_index = max(indel_indices)
    ref_nucs_after = total_reference_nucs(cigar[last_indel_index + 1:])
    return min(ref_nucs_before, ref_nucs_after)

def contains_splicing(read):
    return any(k == BAM_CREF_SKIP for k, l in read.cigar)

def contains_soft_clipping(parsed_line):
    cigar_blocks = cigar_string_to_blocks(parsed_line['CIGAR'])
    kinds = [k for k, l in cigar_blocks]
    return ('S' in kinds)

def contains_soft_clipping_pysam(read):
    kinds = [k for k, l in read.cigar]
    return (BAM_CSOFT_CLIP in kinds)

def get_soft_clipped_block(alignment, edge):
    strand = get_strand(alignment)

    if (edge == 5 and strand == '+') or (edge == 3 and strand == '-'):
        op, length = alignment.cigar[0]
        if op == BAM_CSOFT_CLIP:
            sl = slice(None, length)
        else:
            sl = slice(0)
    elif (edge == 5 and strand == '-') or (edge == 3 and strand == '+'):
        op, length = alignment.cigar[-1]
        if op == BAM_CSOFT_CLIP:
            sl = slice(-length, None)
        else:
            sl = slice(0)

    seq = alignment.seq[sl]
    qual = alignment.query_qualities[sl]

    return seq, qual

def get_max_soft_clipped_length(alignment):
    max_soft_clipped = 0
    for edge in [5, 3]:
        seq, qual = get_soft_clipped_block(alignment, edge)
        max_soft_clipped = max(max_soft_clipped, len(seq))
    return max_soft_clipped

def cigar_blocks_to_string(cigar_blocks):
    ''' Builds a CIGAR string out of a corresponding list of operations. '''
    string = [f'{length}{op_to_char[op]}' for op, length in cigar_blocks]
    return ''.join(string)

def aligned_pairs_to_cigar(aligned_pairs, guide=None):
    op_sequence = []
    for read, ref in aligned_pairs:
        if read == None or read == '-':
            op_sequence.append(BAM_CDEL)
        elif read == 's':
            op_sequence.append(BAM_CREF_SKIP)
        elif ref == None or ref == '-':
            op_sequence.append(BAM_CINS)
        elif ref == 'S':
            op_sequence.append(BAM_CSOFT_CLIP)
        else:
            op_sequence.append(BAM_CMATCH)

    cigar = [(op, len(times)) for op, times in utilities.group_by(op_sequence)]

    if guide:
        guide_cigar, from_side = guide

        if from_side == 'right':
            cigar = cigar[::-1]
            guide_cigar = guide_cigar[::-1]

        for i in range(min(len(cigar), len(guide_cigar))):
            op, length = cigar[i]
            guide_op, guide_length = guide_cigar[i]
            cigar[i] = (guide_op, length)
        
        if from_side == 'right':
            cigar = cigar[::-1]
            guide_cigar = guide_cigar[::-1]

    return cigar

def cigar_to_aligned_pairs(cigar, start):
    aligned_pairs = []

    ref_pos = start
    read_pos = 0
    for op, length in cigar:
        if op == BAM_CMATCH or op == BAM_CEQUAL or op == BAM_CDIFF:
            for i in range(length):
                aligned_pairs.append((read_pos, ref_pos))

                ref_pos += 1
                read_pos += 1

        elif op == BAM_CDEL:
            # Deletion results in gap in read
            for i in range(length):
                aligned_pairs.append((None, ref_pos))
                
                ref_pos += 1
        
        elif op == BAM_CREF_SKIP:
            # Skip results in gap in read
            for i in range(length):
                aligned_pairs.append(('s', ref_pos))
                
                ref_pos += 1

        elif op == BAM_CINS:
            # Insertion results in gap in ref
            for i in range(length):
                aligned_pairs.append((read_pos, None))

                read_pos += 1
        
        elif BAM_CSOFT_CLIP:
            # Soft-clipping results in gap in ref
            for i in range(length):
                aligned_pairs.append((read_pos, 'S'))

                read_pos += 1

        else:
            raise ValueError('Unsupported op', cigar)

    return aligned_pairs

def cigar_to_aligned_pairs_backwards(cigar, end, read_length):
    aligned_pairs = []

    ref_pos = end
    read_pos = read_length - 1
    for op, length in cigar[::-1]:
        if op == BAM_CMATCH or op == BAM_CEQUAL or op == BAM_CDIFF:
            for i in range(length):
                aligned_pairs.append((read_pos, ref_pos))

                ref_pos -= 1
                read_pos -= 1

        elif op == BAM_CDEL or op == BAM_CREF_SKIP:
            # Deletion results in gap in read
            for i in range(length):
                aligned_pairs.append((None, ref_pos))
                
                ref_pos -= 1

        elif op == BAM_CINS:
            # Insertion results in gap in ref
            for i in range(length):
                aligned_pairs.append((read_pos, None))

                read_pos -= 1
        
        elif op == BAM_CSOFT_CLIP:
            # Soft-clipping results in gap in ref
            for i in range(length):
                aligned_pairs.append((read_pos, 'S'))

                read_pos -= 1

        else:
            raise ValueError('Unsupported op', cigar)

    return aligned_pairs

def truncate_cigar_blocks_up_to(cigar_blocks, truncated_length):
    ''' Given pysam-style cigar_blocks, truncates the blocks to explain
        truncated_length read bases.
    '''
    bases_so_far = 0
    truncated_blocks = []

    for operation, length in cigar_blocks:
        # If the next block wouldn't consume anything, we want to include it.
        if bases_so_far == truncated_length and operation in read_consuming_ops:
            break

        if operation in read_consuming_ops:
            length_to_use = min(truncated_length - bases_so_far, length)
            bases_so_far += length_to_use
        else:
            length_to_use = length

        truncated_blocks.append((operation, length_to_use))

        # If we didn't use the whole block, need to break because the next loop
        # will try to use the next block if it is non-consuming.
        if length_to_use < length:
            break

    return truncated_blocks

def truncate_cigar_blocks_from_beginning(cigar_blocks, truncated_length):
    ''' Removes cigar operations from the beginning of cigar_blocks so that
    truncated_length total read bases remain accounted for.
    '''
    flipped_truncated_blocks = truncate_cigar_blocks_up_to(cigar_blocks[::-1],
                                                           truncated_length,
                                                          )
    truncated_blocks = flipped_truncated_blocks[::-1]
    return truncated_blocks

def collapse_cigar_blocks(cigar_blocks):
    collapsed = []
    for kind, blocks in utilities.group_by(cigar_blocks, lambda t: t[0]):
        collapsed.append((kind, sum(t[1] for t in blocks)))
    return collapsed

def aligned_pairs_to_MD_string(aligned_pairs):
    ''' Produce an MD string from an alignment. '''
    # Mark all blocks of matching with numbers, all deletion bases with '^*0', and all mismatch bases.
    MD_list = []
    current_match_length = 0
    for ref_char, read_char in aligned_pairs:
        if ref_char == read_char:
            current_match_length += 1
        elif ref_char != '-':
            if current_match_length > 0:
                MD_list.append(current_match_length)
                current_match_length = 0

            if read_char == '-':
                MD_list.append(0)
            else:
                MD_list.append(ref_char)
    
    if current_match_length > 0:
        MD_list.append(current_match_length)

    # Remove all zeros that aren't a deletion followed by a mismatch
    reduced_MD_list = []
    for i in range(len(MD_list)):
        if isinstance(MD_list[i], int):
            if MD_list[i] > 0:
                reduced_MD_list.append(MD_list[i])
            elif 0 < i < len(MD_list) - 1:
                if isinstance(MD_list[i - 1], str) and isinstance(MD_list[i + 1], str) and MD_list[i - 1][0] == '^' and MD_list[i + 1][0] != '^':
                    reduced_MD_list.append(MD_list[i])
        else:
            reduced_MD_list.append(MD_list[i])
    
    # Collapse all deletions.
    collapsed_MD_list = [reduced_MD_list[0]]
    for i in range(1, len(reduced_MD_list)):
        if isinstance(collapsed_MD_list[-1], str) and collapsed_MD_list[-1][0] == '^' and \
           isinstance(reduced_MD_list[i], str) and reduced_MD_list[i][0] == '^':
            
            collapsed_MD_list[-1] += reduced_MD_list[i][1]
        else:
            collapsed_MD_list.append(reduced_MD_list[i])

    # The standard calls for a number to start and to end, zero if necessary.
    if isinstance(collapsed_MD_list[0], str):
        collapsed_MD_list.insert(0, 0)
    if isinstance(collapsed_MD_list[-1], str):
        collapsed_MD_list.append(0)
    
    MD_string = ''.join(map(str, collapsed_MD_list))
    return MD_string

md_number = re.compile(r'[0-9]+')
md_text = re.compile(r'[A-Z]+')

def md_string_to_ops_string(md_string):
    ''' Converts an MD string into a list of operations for supplying reference
        characters, either '=' if equal to the read, or any other char if equal
        to that char.
    '''
    # In the presence of a CIGAR string, the '^' character seems extraneous.
    # 94 is unicode ^.
    md_string = md_string.translate({94: None})
    
    match_lengths = [int(s) for s in re.findall(md_number, md_string)]
    text_blocks = re.findall(md_text, md_string)
    
    # The standard calls for a number to start and end, zero if necessary,
    # so after removing the initial number, there must be the same number of 
    # match_lengths and text_blocks.
    if len(text_blocks) != len(match_lengths) - 1:
        raise ValueError(md_string)
    
    ops_string = '='*match_lengths[0]
    for text_block, match_length in zip(text_blocks, match_lengths[1:]):
        ops_string += text_block
        ops_string += '='*match_length
    
    return ops_string

md_item_pattern = re.compile(r'[0-9]+|[TCAGN^]+')

def int_if_possible(string):
    try:
        int_value = int(string)
        return int_value
    except ValueError:
        return string

def md_string_to_items(md_string):
    items = [int_if_possible(item) for item in md_item_pattern.findall(md_string)]
    return items

def md_items_to_md_string(items):
    string = ''.join(str(item) for item in items)
    return string

def reverse_md_items(items):
    reversed_items = []
    for item in items[::-1]:
        if isinstance(item, int):
            reversed_items.append(item)
        else:
            if item.startswith('^'):
                reversed_items.append('^' + item[:0:-1])
            else:
                reversed_items.append(item[::-1])
    return reversed_items

def truncate_md_items(md_items, truncated_length):
    if truncated_length == 0:
        truncated_items = [0]
    else:
        bases_so_far = 0
        truncated_items = []

        for item in md_items:
            if bases_so_far == truncated_length:
                break

            if isinstance(item, int):
                length_to_use = min(truncated_length - bases_so_far, item)
                bases_so_far += length_to_use
                truncated_items.append(length_to_use)
            else:
                if item.startswith('^'):
                    truncated_item = ['^']
                    item = item[1:]
                else:
                    truncated_item = []

                for c in item:
                    if c == '^':
                        raise ValueError
                    
                    if bases_so_far == truncated_length:
                        break

                    truncated_item.append(c)
                    bases_so_far += 1

                truncated_items.append(''.join(truncated_item))

        # Ensure that it starts and ends with a number
        if not isinstance(truncated_items[0], int):
            truncated_items = [0] + truncated_items
        if not isinstance(truncated_items[-1], int):
            truncated_items = truncated_items + [0]

    return truncated_items

def combine_md_strings(first_string, second_string):
    if first_string == '':
        return second_string
    if second_string == '':
        return first_string

    first_items = md_string_to_items(first_string)
    second_items = md_string_to_items(second_string)
    before = first_items[:-1]
    after = second_items[1:]

    if isinstance(first_items[-1], int):
        if isinstance(second_items[0], int):
            interface = [first_items[-1] + second_items[0]]
        else:
            interface = [first_items[-1], second_items[0]]
    else:
        if isinstance(second_items[0], int):
            interface = [first_items[-1], second_items[0]]
        else:
            if first_items[-1].startswith('^'):
                if second_items[0].startswith('^'):
                    interface = [first_items[-1] + second_items[0][1:]]
                else:
                    interface = [first_items[-1], 0, second_items[0]]
            else:
                if second_items[0].startswith('^'):
                    interface = [first_items[-1], 0, second_items[0]]
                else:
                    interface = [first_items[-1] + second_items[0]]

    combined_items = before + interface + after

    combined_string = md_items_to_md_string(combined_items)
    
    return combined_string

def truncate_md_string_up_to(md_string, truncated_length):
    ''' Truncates from the end of md_string so that the result only consumes
    truncated_length ref characters.
    '''
    md_items = md_string_to_items(md_string)
    truncated_items = truncate_md_items(md_items, truncated_length)
    return md_items_to_md_string(truncated_items)

def truncate_md_string_from_beginning(md_string, truncated_length):
    ''' Truncates from the beginning of md_string so that the result only
    consumes truncated_length ref characters.
    '''
    md_items = md_string_to_items(md_string)
    reversed_items = reverse_md_items(md_items)
    reversed_truncated_items = truncate_md_items(reversed_items, truncated_length)
    truncated_items = reverse_md_items(reversed_truncated_items)
    return md_items_to_md_string(truncated_items)
    
def produce_alignment(mapping):
    ''' Returns a list of (ref_char, read_char, qual_char, ref_pos, read_pos)
        tuples.
    '''
    read_seq = mapping.seq
    if read_seq == None:
        read_seq = ''
    
    read_quals = mapping.query_qualities
    if read_quals == None:
        read_quals = []
    
    MD_string = dict(mapping.tags)['MD']
    
    ref_ops = iter(md_string_to_ops_string(MD_string))
    
    columns = []

    ref_pos = mapping.pos
    read_pos = 0
    for op, length in mapping.cigar:
        if op == BAM_CMATCH or op == BAM_CEQUAL or op == BAM_CDIFF:
            for i in range(length):
                read_char = read_seq[read_pos]
                
                ref_op = next(ref_ops)
                if ref_op == '=':
                    ref_char = read_char
                else:
                    ref_char = ref_op
                
                qual = read_quals[read_pos]
                
                column = (ref_char, read_char, qual, ref_pos, read_pos)
                columns.append(column)

                ref_pos += 1
                read_pos += 1

        elif op == BAM_CDEL:
            # Deletion results in gap in read
            for i in range(length):
                read_char = '-'
                ref_char = next(ref_ops)
                qual = 0 
                
                column = (ref_char, read_char, qual, ref_pos, read_pos)
                columns.append(column)
                
                ref_pos += 1

        elif op == BAM_CINS:
            # Insertion results in gap in ref
            for i in range(length):
                read_char = read_seq[read_pos]
                ref_char = '-'
                qual = read_quals[read_pos]
                column = (ref_char, read_char, qual, ref_pos, read_pos)
                columns.append(column)

                read_pos += 1

        elif op == BAM_CREF_SKIP:
            ref_pos += length

        elif op == BAM_CSOFT_CLIP:
            read_pos += length
    
    return columns

def ref_dict_from_mapping(mapping):
    ''' Build a dictionary mapping reference positions to base identities from
    the cigar and MD tag of a mapping.
    '''
    alignment = produce_alignment(mapping)
    ref_dict = {}
    for ref_char, _, _, ref_position, _ in alignment:
        if ref_char == '-':
            continue

        if ref_position in ref_dict:
            # A ref_position shouldn't appear more than once
            raise ValueError(mapping)

        ref_dict[ref_position] = ref_char

    return ref_dict

def merge_ref_dicts(first_dict, second_dict):
    ''' Merge dictionaries mapping reference positions to base identities. '''
    merged_dict = {}
    merged_dict.update(first_dict)
    for position, base in second_dict.items():
        if position in merged_dict:
            if merged_dict[position] != base:
                # contradiction
                raise ValueError(first_dict, second_dict)
        else:
            merged_dict[position] = base

    return merged_dict

def sort_bam(input_file_name, output_file_name, by_name=False, num_threads=1):
    output_file_name = Path(output_file_name)

    samtools_command = ['samtools', 'sort']
    if by_name:
        samtools_command.append('-n')

    # For unambiguously marking temp files from this function.
    tail = 'SORT_TEMP'

    # Clean up any temporary files left behind by previous attempts to sort.
    pattern = f'{output_file_name.name}\.{tail}\.\d\d\d\d\.bam'
    for fn in output_file_name.parent.iterdir():
        if re.match(pattern, fn.name):
            fn.unlink()

    samtools_command.extend(['-@', str(num_threads),
                             '-T', str(output_file_name) + f'.{tail}',
                             '-o', str(output_file_name),
                             str(input_file_name),
                            ])

    try:
        subprocess.run(samtools_command, check=True, stderr=subprocess.PIPE)
    except subprocess.CalledProcessError as e:
        print(e.stderr)
        raise

    if not by_name:
        pysam.index(str(output_file_name))

def merge_sorted_bam_files(input_file_names, merged_file_name, by_name=False, make_index=True):
    # To avoid running into max open file limits, split into groups of 500.
    if len(input_file_names) > 500:
        chunks = utilities.list_chunks(input_file_names, 500)
        merged_chunk_fns = []
        for i, chunk in enumerate(chunks):
            merged_chunk_fn = f'{merged_file_name}.intermediate_chunk.{i:04d}'
            merged_chunk_fns.append(merged_chunk_fn)
            merge_sorted_bam_files(chunk, merged_chunk_fn, by_name=by_name, make_index=False)

        merge_sorted_bam_files(merged_chunk_fns, merged_file_name, by_name=by_name)

        for merged_chunk_fn in merged_chunk_fns:
            os.remove(merged_chunk_fn)

    else:
        input_file_names = [str(fn) for fn in input_file_names]
        merged_file_name = str(merged_file_name)

        if len(input_file_names) == 1:
            shutil.copy(input_file_names[0], merged_file_name)
        else:
            merge_command = ['samtools', 'merge', '-f']

            if by_name:
                merge_command.append('-n')

            merge_command.extend([merged_file_name] + input_file_names)

            try:
                subprocess.run(merge_command, check=True, stderr=subprocess.PIPE)
            except subprocess.CalledProcessError as e:
                print(e.stderr)
                raise
        
        if make_index and not by_name:
            try:
                pysam.index(merged_file_name)
            except pysam.utils.SamtoolsError:
                # Need to sort the merged file because at least one input file was missing a target.
                temp_sorted_name = merged_file_name + '.temp_sorted'
                sort_bam(merged_file_name, temp_sorted_name)
                os.rename(temp_sorted_name, merged_file_name)
                os.rename(temp_sorted_name + '.bai', merged_file_name + '.bai')

def bam_to_sam(bam_file_name, sam_file_name):
    view_command = ['samtools', 'view', '-h', '-o', sam_file_name, bam_file_name]
    subprocess.check_call(view_command)

def get_length_counts(bam_file_name, only_primary=True, only_unique=False):
    bam_file = pysam.AlignmentFile(bam_file_name)
    if only_unique:
        qlen_counts = Counter(ar.qlen for ar in bam_file if ar.mapping_quality == 50)
    elif only_primary:
        qlen_counts = Counter(ar.qlen for ar in bam_file if not ar.is_unmapped and not ar.is_secondary)
    else:
        qlen_counts = Counter(ar.qlen for ar in bam_file)
    
    return qlen_counts

def get_tlen_counts(bam_file_name, only_primary=True, only_unique=False):
    bam_file = pysam.AlignmentFile(bam_file_name)
    if only_unique:
        tlen_counts = Counter(ar.tlen for ar in bam_file if ar.mapping_quality == 50)
    elif only_primary:
        tlen_counts = Counter(ar.tlen for ar in bam_file if not ar.is_unmapped and not ar.is_secondary)
    else:
        tlen_counts = Counter(ar.tlen for ar in bam_file)
    
    return tlen_counts

def get_mapq_counts(bam_file_name):
    bam_file = pysam.AlignmentFile(bam_file_name)
    mapq_counts = Counter(ar.mapq for ar in bam_file)
    return mapq_counts

def mapping_to_Read(mapping):
    seq = mapping.get_forward_sequence()
    qual = fastq.encode_sanger(mapping.get_forward_qualities())

    read = fastq.Read(mapping.query_name, seq, qual)
    return read

def sam_to_fastq(sam_file_name):
    sam_file = pysam.AlignmentFile(str(sam_file_name))
    for mapping in sam_file:
        yield mapping_to_Read(mapping)

bam_to_fastq = sam_to_fastq

class AlignmentSorter:
    ''' Context manager that handles writing AlignedSegments into a samtools
    sort process.
    '''
    temp_prefix_tail = 'ALIGNMENTSORTER_TEMP'

    def __init__(self, output_file_name, header, by_name=False):
        self.header = header
        self.output_file_name = Path(output_file_name)
        self.by_name = by_name
        self.fifo = mapping_tools.TemporaryFifo(name='unsorted_fifo.bam')

    def remove_temporary_files(self):
        ''' Find any temporary files that might have been left behind by a
        previous call with the same output_file_name.
        '''
        pattern = f'{self.output_file_name.name}\.{AlignmentSorter.temp_prefix_tail}\.\d\d\d\d\.bam'
        for fn in self.output_file_name.parent.iterdir():
            if re.match(pattern, fn.name):
                fn.unlink()

    def __enter__(self):
        self.fifo.__enter__()

        self.remove_temporary_files()

        sort_command = ['samtools', 'sort']
        if self.by_name:
            sort_command.append('-n')

        self.dev_null = open(os.devnull, 'w')
        sort_command.extend(['-T', str(self.output_file_name) + f'.{AlignmentSorter.temp_prefix_tail}',
                             '-o', str(self.output_file_name),
                             self.fifo.file_name,
                            ])
        self.sort_process = subprocess.Popen(sort_command,
                                             stderr=subprocess.PIPE,
                                            )

        self.sam_file = pysam.AlignmentFile(self.fifo.file_name, 'wbu', header=self.header)

        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.sam_file.close()
        _, err_output = self.sort_process.communicate()
        self.dev_null.close()
        self.fifo.__exit__(exception_type, exception_value, exception_traceback)
        
        self.remove_temporary_files()

        if self.sort_process.returncode:
            raise RuntimeError(err_output)

        if not self.by_name:
            pysam.index(str(self.output_file_name))

    def write(self, alignment):
        self.sam_file.write(alignment)

class multiple_AlignmentSorters(contextlib.ExitStack):
    def __init__(self, header=None, by_name=False):
        super().__init__()
        self.sorters = {}
        self.header = header
        self.by_name = by_name
            
    def __enter__(self):
        super().__enter__()
        for name in self.sorters:
            self.enter_context(self.sorters[name])

        return self
            
    def __getitem__(self, key):
        return self.sorters[key]

    def __setitem__(self, name, fn_and_possibly_header):
        if isinstance(fn_and_possibly_header, tuple):
            fn, header = fn_and_possibly_header
        else:
            fn = fn_and_possibly_header
            header = self.header

        self.sorters[name] = AlignmentSorter(fn, header, self.by_name)

class AlignedSegmentByName(object):
    def __init__(self, aligned_segment):
        self.aligned_segment = aligned_segment
        
    def __lt__(self, other):
        return self.aligned_segment.query_name < other.aligned_segment.query_name

def merge_by_name(*mapping_iterators):
    ''' Merges iterators over mappings that are sorted by name.
    '''
    wrapped_iterators = [(AlignedSegmentByName(m) for m in mappings) for mappings in mapping_iterators]
    merged_wrapped = heapq.merge(*wrapped_iterators)
    last_qname = None
    for al_by_name in merged_wrapped:
        qname = al_by_name.aligned_segment.query_name
        if last_qname is not None and qname < last_qname:
            print(last_qname, qname)
            raise ValueError('Attempted to merge unsorted mapping iterators')

        last_qname = qname
        yield al_by_name.aligned_segment

def aligned_pairs_exclude_soft_clipping(mapping):
    cigar = mapping.cigartuples
    aligned_pairs = mapping.aligned_pairs
    
    first_op, first_length = cigar[0]

    if first_op == BAM_CSOFT_CLIP:
        aligned_pairs = aligned_pairs[first_length:]

    if len(cigar) > 1:
        last_op, last_length = cigar[-1]
        if last_op == BAM_CSOFT_CLIP and last_length != 0:
            aligned_pairs = aligned_pairs[:-last_length]

    return aligned_pairs

def parse_idxstats(bam_fn):
    lines = pysam.idxstats(str(bam_fn)).splitlines()
    fields = [line.split('\t') for line in lines]
    parsed = {rname: int(count) for rname, _, count, _ in fields}
    return parsed

def get_num_alignments(bam_fn):
    return sum(parse_idxstats(bam_fn).values())

def collapse_soft_clip_blocks(cigar_blocks):
    ''' If there are multiple consecutive soft clip blocks on either end,
    collapse them into a single block.
    '''
    first_non_cigar_index = 0
    while cigar_blocks[first_non_cigar_index][0] == BAM_CSOFT_CLIP:
        first_non_cigar_index += 1

    if first_non_cigar_index > 0:
        total_length = sum(length for kind, length in cigar_blocks[:first_non_cigar_index])
        if total_length > 0:
            to_add = [(BAM_CSOFT_CLIP, total_length)]
        else:
            to_add = []
        cigar_blocks = to_add + cigar_blocks[first_non_cigar_index:]
    
    last_non_cigar_index = len(cigar_blocks) - 1
    while cigar_blocks[last_non_cigar_index][0] == BAM_CSOFT_CLIP:
        last_non_cigar_index -= 1

    if last_non_cigar_index < len(cigar_blocks) - 1:
        total_length = sum(length for kind, length in cigar_blocks[last_non_cigar_index + 1:])
        if total_length > 0:
            to_add = [(BAM_CSOFT_CLIP, total_length)]
        else:
            to_add = []
        cigar_blocks = cigar_blocks[:last_non_cigar_index + 1] + to_add

    return cigar_blocks

def crop_al_to_query_int(alignment, start, end):
    ''' Replace any parts of alignment that involve query bases not in the
    interval [start, end] with soft clipping.
    query coords are given relative to the original read (and are therefore
    transformed if alignment is reversed.)
    '''
    alignment = copy.deepcopy(alignment)

    if alignment is None or alignment.is_unmapped:
        return alignment
    
    if end < start:
        # query interval is empty
        return None
    else:
        overlap = interval.Interval(start, end) & interval.get_covered(alignment)
        if len(overlap) == 0:
            return None

    if alignment.is_reverse:
        start, end = alignment.query_length - 1 - end, alignment.query_length - 1 - start

    aligned_pairs = cigar_to_aligned_pairs(alignment.cigar, alignment.reference_start)

    start_i = 0
    read, ref = aligned_pairs[start_i]
    while read is None or read == 's' or read < start:
        start_i += 1
        read, ref = aligned_pairs[start_i]
        
    end_i = len(aligned_pairs) - 1
    read, ref = aligned_pairs[end_i]
    while read is None or read == 's' or read > end:
        end_i -= 1
        read, ref = aligned_pairs[end_i]

    if alignment.has_tag('MD'):
        MD = alignment.get_tag('MD')

        total_ref_nucs = total_reference_nucs_except_splicing(alignment.cigar)
        removed_from_start = total_reference_nucs_except_splicing(aligned_pairs_to_cigar(aligned_pairs[:start_i]))
        removed_from_end = total_reference_nucs_except_splicing(aligned_pairs_to_cigar(aligned_pairs[end_i + 1:]))

        MD = truncate_md_string_up_to(MD, total_ref_nucs - removed_from_end)
        MD = truncate_md_string_from_beginning(MD, total_ref_nucs - removed_from_end - removed_from_start)

        alignment.set_tag('MD', MD)

    restricted_pairs = aligned_pairs[start_i:end_i + 1]
    cigar = aligned_pairs_to_cigar(restricted_pairs) 

    before_soft = start
    if before_soft > 0:
        cigar = [(BAM_CSOFT_CLIP, before_soft)] + cigar

    after_soft = alignment.query_length - end - 1
    if after_soft > 0:
        cigar = cigar + [(BAM_CSOFT_CLIP, after_soft)]

    cigar = collapse_soft_clip_blocks(cigar)

    restricted_rs = [r for q, r in restricted_pairs if r != None and r != 'S']
    if not restricted_rs:
        alignment.is_unmapped = True
        alignment.cigar = []
    else:
        alignment.reference_start = min(restricted_rs)
        alignment.cigar = cigar

    return alignment

def crop_al_to_ref_int(alignment, start, end):
    ''' Returns a copy of alignment in which any query bases that align
    outside the interval [start, end] are soft-clipped. If no bases are left,
    sets alignment.is_unmapped to true.
    '''
    if alignment is None:
        return None

    original_alignment = alignment
    alignment = copy.deepcopy(alignment)

    if alignment.reference_start > end or alignment.reference_end - 1 < start:
        # alignment doesn't overlap the ref interval at all
        return None

    if alignment.reference_start >= start and alignment.reference_end - 1 <= end:
        # alignment is entirely contained in the ref_interval
        return alignment

    query_length = alignment.query_length
    aligned_pairs = cigar_to_aligned_pairs(alignment.cigar, alignment.reference_start)

    start_i = 0
    while (
        aligned_pairs[start_i][1] == 'S' or
        aligned_pairs[start_i][0] is None or
        aligned_pairs[start_i][1] is None or
        aligned_pairs[start_i][1] < start
    ):
        start_i += 1
        
    end_i = len(aligned_pairs) - 1
    while (
        aligned_pairs[end_i][1] == 'S' or
        aligned_pairs[end_i][0] is None or
        aligned_pairs[end_i][1] is None or
        aligned_pairs[end_i][1] > end
    ):
        end_i -= 1

    remaining = aligned_pairs[start_i:end_i + 1]
    if remaining:
        cigar = aligned_pairs_to_cigar(remaining)
        before_soft = remaining[0][0]
        if before_soft > 0:
            cigar = [(BAM_CSOFT_CLIP, before_soft)] + cigar

        after_soft = query_length - remaining[-1][0] - 1
        if after_soft > 0:
            cigar = cigar + [(BAM_CSOFT_CLIP, after_soft)]

        alignment.cigar = cigar

        alignment.reference_start = aligned_pairs[start_i][1]
    else:
        alignment.is_unmapped = True
        alignment.cigar = []

    if alignment.has_tag('MD'):
        MD = alignment.get_tag('MD')

        total_ref_nucs = total_reference_nucs_except_splicing(original_alignment.cigar)
        removed_from_start = total_reference_nucs_except_splicing(aligned_pairs_to_cigar(aligned_pairs[:start_i]))
        removed_from_end = total_reference_nucs_except_splicing(aligned_pairs_to_cigar(aligned_pairs[end_i + 1:]))

        MD = truncate_md_string_up_to(MD, total_ref_nucs - removed_from_end)
        MD = truncate_md_string_from_beginning(MD, total_ref_nucs - removed_from_end - removed_from_start)

        alignment.set_tag('MD', MD)
    
    return alignment

def crop_al_to_feature(al, feature):
    cropped_al = crop_al_to_ref_int(al, feature.start, feature.end)
    return cropped_al

def query_interval_overlapping_feature(al, feature):
    cropped_al = crop_al_to_feature(al, feature)
    return interval.get_covered(cropped_al)

def disallow_query_positions_from_other(alignment, other):
    start, end = query_interval(alignment)
    other_start, other_end = query_interval(other)
    if other_start <= end or other_end >= start:
        if other_start > start and other_end < end:
            raise ValueError
        elif other_start <= start:
            alignment = crop_al_to_query_int(alignment, other_end + 1, alignment.query_length - 1)
        elif other_end >= end:
            alignment = crop_al_to_query_int(alignment, 0, other_start - 1)

    return alignment

def query_interval(alignment):
    start = alignment.query_alignment_start
    end = alignment.query_alignment_end - 1

    if alignment.is_reverse:
        start, end = true_query_position(end, alignment), true_query_position(start, alignment)

    return start, end

def ref_interval(alignment):
    return alignment.reference_start, alignment.reference_end - 1

def merge_multiple_adjacent_alignments(als, ref_seqs):
    merger = functools.partial(merge_adjacent_alignments, ref_seqs=ref_seqs)
    als = sorted(als, key=query_interval)
    return functools.reduce(merger, als)

def merge_any_adjacent_pairs(als, ref_seqs):
    als = sorted(als, key=lambda al: interval.get_covered(al))

    if len(als) == 0:
        return als
    else:
        merged_als = [als[0]]

        while len(als) > 0:
            left_al = merged_als.pop()
            right_al = als.pop(0)
            
            merged = merge_adjacent_alignments(left_al, right_al, ref_seqs)
            if merged is not None:
                merged_als.append(merged)
            else:
                merged_als.extend([left_al, right_al])

        return merged_als

def merge_adjacent_alignments(first, second, ref_seqs):
    ''' If first and second are alignments to the same reference name and strand
    that are adjacent or partially overlap on the query, returns a single merged
    alignment with an appropriately sized deletion that minimizes edit distance,
    otherwise return None.
    '''
    if first is None or second is None:
        return None

    if first == second:
        return first

    if first.reference_name != second.reference_name:
        return None
    else:
        ref_seq = ref_seqs[first.reference_name]

    if get_strand(first) != get_strand(second):
        return None

    left_query, right_query = sorted([first, second], key=query_interval)
    left_covered = interval.get_covered(left_query)
    right_covered = interval.get_covered(right_query)

    # Ensure that the alignments point towards each other.
    strand = get_strand(first)
    if strand == '+':
        left_cropped = crop_al_to_query_int(left_query, 0, right_covered.start - 1)
        if left_cropped is None or left_cropped.is_unmapped:
            # left alignment doesn't cover any query not covered by right
            return None

        if left_cropped.reference_end > right_query.reference_start:
            return None

    elif strand == '-':
        right_cropped = crop_al_to_query_int(right_query, left_covered.end + 1, np.inf)
        if right_cropped is None or right_cropped.is_unmapped:
           # right alignment doesn't cover any query not covered by left
           return None

        if right_cropped.reference_end > left_query.reference_start:
            return None

    if interval.are_adjacent(left_covered, right_covered):
        left_cropped, right_cropped = left_query, right_query

    elif interval.are_disjoint(left_covered, right_covered):
        return None

    else:
        overlap = left_covered & right_covered
        left_ceds = cumulative_edit_distances(left_query, overlap, False, ref_seq=ref_seq)
        right_ceds = cumulative_edit_distances(right_query, overlap, True, ref_seq=ref_seq)

        switch_after_edits = {
            overlap.start - 1 : right_ceds[overlap.start],
            overlap.end: left_ceds[overlap.end],
        }

        for q in range(overlap.start, overlap.end):
            switch_after_edits[q] = left_ceds[q] + right_ceds[q + 1]

        min_edits = min(switch_after_edits.values())
        best_switch_points = [s for s, d in switch_after_edits.items() if d == min_edits]
        switch_after = best_switch_points[0]

        left_cropped = crop_al_to_query_int(left_query, 0, switch_after)
        right_cropped = crop_al_to_query_int(right_query, switch_after + 1, right_query.query_length)

    if left_cropped is None or left_cropped.is_unmapped or right_cropped is None or right_cropped.is_unmapped:
        # this may not be appropriate in all circumstances
        return None

    left_ref, right_ref = sorted([left_cropped, right_cropped], key=lambda al: al.reference_start)

    if left_ref.reference_end > right_ref.reference_start:
        return None

    deletion_length = right_ref.reference_start - left_ref.reference_end
    if deletion_length > 0:
        deletion_cigar = [(BAM_CDEL, deletion_length)]
    else:
        deletion_cigar = []

    merged = copy.deepcopy(left_ref)
    merged.cigar = left_ref.cigar[:-1] + deletion_cigar + right_ref.cigar[1:]

    return merged

def cumulative_edit_distances(mapping, query_interval, from_end, ref_seq=None):
    ''' Returns a dictionary of how many cumulatives edits are involved
    in mapping from the beginning (or end, if from_end is True) of query_interval
    to each query position in query_interval.
    '''
    tuples = aligned_tuples(mapping, ref_seq=ref_seq)
    
    if get_strand(mapping) == '-':    
        tuples = tuples[::-1]
        
    beginning = [i for i, (q, read, r, ref, qual) in enumerate(tuples) if q == query_interval.start][0]
    end = [i for i, (q, read, r, ref, qual) in enumerate(tuples) if q == query_interval.end][-1]
    
    relevant = tuples[beginning:end + 1]
    if from_end:
        relevant = relevant[::-1]

    c_e_ds = {}

    total = 0

    for q, read_b, r, ref_b, qual in relevant:
        if read_b != ref_b:
            total += 1
        
        if q is not None:
            c_e_ds[q] = total
            
    return c_e_ds

def find_best_query_switch_after(left_al, right_al, left_ref_seq, right_ref_seq, tie_break):
    ''' If left_al and right_al overlap on the query, find the query position such that switching from
    left_al to right_al after that position minimizes the total number of edits.
    '''
    left_covered = interval.get_covered(left_al)
    right_covered = interval.get_covered(right_al)
    overlap = left_covered & right_covered

    if left_al is None:
        if right_al is None:
            raise ValueError
        gap_interval = interval.Interval(0, right_covered.start - 1)
    elif right_al is None:
        if left_al is None:
            raise ValueError
        gap_interval = interval.Interval(left_covered.end + 1, len(left_al.seq) - 1)
    else:
        gap_interval = interval.Interval(left_covered.end + 1, right_covered.start - 1)

    if overlap:
        left_ceds = cumulative_edit_distances(left_al, overlap, False, ref_seq=left_ref_seq)
        right_ceds = cumulative_edit_distances(right_al, overlap, True, ref_seq=right_ref_seq)

        switch_after_edits = {
            overlap.start - 1: right_ceds[overlap.start],
            overlap.end: left_ceds[overlap.end],
        }

        # 22.08.17: I understand this less well than I thought I did.
        # It's not clear what the behavior should be for entirely contained als,
        # but probably should be that switch_after entirely outside should be
        # included, and if this code is included, this doesn't happen. 
        # # If left_al doesn't covers the query base before overlap.start,
        # # overlap.start - 1 needs to pay a penalty of 1. 
        # if left_covered.start > overlap.start - 1:
        #     switch_after_edits[overlap.start - 1] += 1

        # 22.08.15: does equivalent logic need to be added to overlap.end?

        for q in range(overlap.start, overlap.end):
            switch_after_edits[q] = left_ceds[q] + right_ceds[q + 1]

        min_edits = min(switch_after_edits.values())
        best_switch_points = [s for s, d in switch_after_edits.items() if d == min_edits]
        switch_after = tie_break(best_switch_points)
    else:
        min_edits = 0
        switch_after = left_covered.end
        switch_after_edits = {}
        left_ceds = None
        right_ceds = None
        best_switch_points = [switch_after]

    if gap_interval.is_empty:
        gap_length = 0
        gap_interval = None
    else:
        gap_length = len(gap_interval)
    
    results = {
        'switch_after': switch_after,
        'best_switch_points': sorted(best_switch_points),
        'min_edits': min_edits,
        'gap_interval': gap_interval,
        'gap_length': gap_length,
    }

    return results

def true_query_position(p, alignment):
    if alignment.is_reverse:
        p = alignment.query_length - 1 - p
    return p

def closest_query_position(r, alignment, which_side='either'):
    ''' Return query position paired with r (or with the closest to r) '''
    r_to_q = {r: true_query_position(q, alignment)
              for q, r in alignment.aligned_pairs
              if r is not None and q is not None
             }
    if r in r_to_q:
        q = r_to_q[r]
    else:
        if which_side == 'either':
            rs = r_to_q
        elif which_side == 'before':
            rs = (other_r for other_r in r_to_q if other_r < r)
        elif which_side == 'after':
            rs = (other_r for other_r in r_to_q if other_r > r)

        closest_r = min(rs, key=lambda other_r: abs(other_r - r))
        q = r_to_q[closest_r]

    return q

def closest_ref_position(q, alignment, which_side='either'):
    ''' Return ref position paired with q (or with the closest to q) '''
    q_to_r = {true_query_position(q, alignment): r
              for q, r in alignment.aligned_pairs
              if r is not None and q is not None
             }

    if q in q_to_r:
        r = q_to_r[q]
    else:
        if which_side == 'either':
            qs = q_to_r
        elif which_side == 'before':
            qs = [other_q for other_q in q_to_r if other_q < q]
        elif which_side == 'after':
            qs = [other_q for other_q in q_to_r if other_q > q]
        elif which_side == 'exact':
            qs = []
        else:
            raise ValueError(which_side)

        if len(qs) == 0:
            r = None
        else:
            closest_q = min(qs, key=lambda other_q: abs(other_q - q))
            r = q_to_r[closest_q]

    return r

def max_block_length(alignment, block_types):
    if alignment is None or alignment.is_unmapped:
        return 0
    else:
        block_lengths = [l for k, l in alignment.cigar if k in block_types]
        if len(block_lengths) == 0:
            return 0
        else:
            return max(block_lengths)

def total_indel_lengths(alignment):
    if alignment.is_unmapped:
        return 0
    else:
        return sum(l for k, l in alignment.cigar if k == BAM_CDEL or k == BAM_CINS)

def get_ref_pos_to_block(alignment):
    ref_pos_to_block = {}
    ref_pos = alignment.reference_start
    for kind, length in alignment.cigar:
        if kind in ref_consuming_ops:
            starts_at = ref_pos

            for r in range(ref_pos, ref_pos + length):
                ref_pos_to_block[r] = (kind, length, starts_at)

            ref_pos += length

    return ref_pos_to_block

def split_at_first_large_insertion(alignment, min_length, exempt_if_overlaps=None):
    q = 0

    # If the alignment is reversed, alignment.cigar is reversed relative to
    # true query positions.
    cigar = alignment.cigar
    if alignment.is_reverse:
        cigar = cigar[::-1]

    # Using cigar blocks, march from beginning of the read to the (possible)
    # insertion point to determine the query interval to crop to.
    for kind, length in cigar:
        if kind == BAM_CINS and length >= min_length:
            before = crop_al_to_query_int(alignment, 0, q - 1)
            after = crop_al_to_query_int(alignment, q + length, alignment.query_length)
            
            if get_strand(alignment) == '+':
                inserted_at = after.reference_start - 0.5
            else:
                inserted_at = before.reference_start - 0.5

            if exempt_if_overlaps is None or inserted_at not in exempt_if_overlaps:
                return [before, after]

        if kind in read_consuming_ops:
            q += length

    return [alignment]

def split_at_large_insertions(alignment, min_length, exempt_if_overlaps=None):
    ''' O(n^2) behavior can be bad for pacbio alignments. '''
    all_split = []
    
    to_split = [alignment]

    while len(to_split) > 0:
        candidate = to_split.pop()
        split = split_at_first_large_insertion(candidate, min_length, exempt_if_overlaps=exempt_if_overlaps)
        if len(split) > 1:
            to_split.extend(split)
        else:
            all_split.extend(split)

    return all_split

def split_at_deletions(alignment, min_length, exempt_if_overlaps=None):
    ''' Split at deletions at least min_length that don't overlap exempt_if_overlaps. '''

    ref_start = alignment.reference_start
    query_bases_before = 0
    query_bases_after = alignment.query_length
    
    tuples = []

    split_at = []
    
    for i, (kind, length) in enumerate(alignment.cigar):
        if kind == BAM_CDEL:
            del_interval = interval.Interval(ref_start, ref_start + length - 1)

            if exempt_if_overlaps is None:
                overlaps = False
            else:
                overlaps = len(del_interval & exempt_if_overlaps) > 0

            if length >= min_length and not overlaps:
                if i != 0 and i != len(alignment.cigar) - 1:
                    split_at.append(i)
        
        if kind in read_consuming_ops:
            read_consumed = length
        else:
            read_consumed = 0
            
        if kind in ref_consuming_ops:
            ref_consumed = length
        else:
            ref_consumed = 0
            
        query_bases_after -= read_consumed
        tuples.append((query_bases_before, query_bases_after, ref_start))
        
        query_bases_before += read_consumed
        ref_start += ref_consumed
        
    split_alignments = []

    if len(split_at) == 0:
        split_alignments = [alignment]
    else:
        split_at = [-1] + split_at + [len(alignment.cigar)]
    
        for i in range(len(split_at) - 1):
            query_bases_before, _, ref_start = tuples[split_at[i] + 1]
            if split_at[i + 1] == len(alignment.cigar):
                query_bases_after = 0
            else:
                _, query_bases_after, _ = tuples[split_at[i + 1]]

            new_cigar = alignment.cigar[split_at[i] + 1:split_at[i + 1]]

            if query_bases_before > 0:
                new_cigar = [(BAM_CSOFT_CLIP, query_bases_before)] + new_cigar

            if query_bases_after > 0:
                new_cigar = new_cigar + [(BAM_CSOFT_CLIP, query_bases_after)]

            split_al = copy.deepcopy(alignment)
            split_al.cigar = new_cigar
            split_al.reference_start = ref_start
            
            split_alignments.append(split_al)

    split_alignments = [soft_clip_terminal_insertions(al) for al in split_alignments]

    return split_alignments

def soft_clip_terminal_insertions(al):
    ''' If al starts or ends with insertions, convert the relevant bases into soft-clipping. '''
    initial_cigar = al.cigar
    clip_lengths = {}
    
    if initial_cigar[0][0] == BAM_CSOFT_CLIP:
        clip_lengths['beginning'] = initial_cigar[0][1]
    else:
        clip_lengths['beginning'] = 0
    
    if len(initial_cigar) > 1 and initial_cigar[-1][0] == BAM_CSOFT_CLIP:
        clip_lengths['end'] = initial_cigar[-1][1]
    else:
        clip_lengths['end'] = 0
        
    non_soft_clipped_blocks = [(kind, length) for kind, length in al.cigar if kind != BAM_CSOFT_CLIP]
    
    if len(non_soft_clipped_blocks) == 0:
        return al
    
    had_terminal_insertion = False
    
    first_kind, first_length = non_soft_clipped_blocks[0]

    if first_kind == BAM_CINS:
        had_terminal_insertion = True
        clip_lengths['beginning'] += first_length

        non_soft_clipped_blocks = non_soft_clipped_blocks[1:]
        
    if len(non_soft_clipped_blocks) > 0:
        last_kind, last_length = non_soft_clipped_blocks[-1]
        
        if last_kind == BAM_CINS:
            had_terminal_insertion = True
            clip_lengths['end'] += last_length

            non_soft_clipped_blocks = non_soft_clipped_blocks[:-1]
    
    if had_terminal_insertion:
        new_cigar = non_soft_clipped_blocks

        if clip_lengths['beginning'] > 0:
            new_cigar = [(BAM_CSOFT_CLIP, clip_lengths['beginning'])] + new_cigar

        if clip_lengths['end'] > 0:
            new_cigar = new_cigar + [(BAM_CSOFT_CLIP, clip_lengths['end'])]

        new_al = copy.deepcopy(al)
        new_al.cigar = new_cigar
    else:
        new_al = al
    
    return new_al

def remove_terminal_deletions(al):
    al = copy.deepcopy(al)
    
    last_block_type, last_block_length = al.cigar[-1]
    if last_block_type == BAM_CDEL:
        al.cigar = al.cigar[:-1]
        if al.is_reverse:
            al.reference_start = al.reference_start + last_block_length
            
    first_block_type, first_block_length = al.cigar[0]
    if first_block_type == BAM_CDEL:
        al.cigar = al.cigar[1:]
        if not al.is_reverse:
            al.reference_start = al.reference_start + first_block_length
    
    return al

def grouped_by_name(als):
    if isinstance(als, (str, Path)):
        saved_verbosity = pysam.set_verbosity(0)
        als = pysam.AlignmentFile(als)
        pysam.set_verbosity(saved_verbosity)

    grouped = utilities.group_by(als, lambda al: al.query_name)

    return grouped

def header_from_STAR_index(index):
    index = Path(index)
    names = (index / 'chrName.txt').read_text().splitlines()
    lengths = [int(l) for l in (index / 'chrLength.txt').read_text().splitlines()]
    header = pysam.AlignmentHeader.from_references(names, lengths)
    return header

def header_from_fasta(fasta_fn):
    fai = fasta.load_fai(fasta_fn).sort_index()

    names = [name for name, row in fai.iterrows()]
    lengths = [row['LENGTH'] for name, row in fai.iterrows()]

    header = pysam.AlignmentHeader.from_references(names, lengths)

    return header

def overlaps_feature(alignment, feature, require_same_strand=True):
    if alignment is None or alignment.is_unmapped:
        return False

    same_reference = alignment.reference_name == feature.seqname
    num_overlapping_bases = feature_overlap_length(alignment, feature)

    if require_same_strand:
        same_strand = (get_strand(alignment) == feature.strand)
    else:
        same_strand = True

    return same_reference and same_strand and (num_overlapping_bases > 0) 

def feature_overlap_length(alignment, feature):
    if alignment is None:
        overlap = 0
    else:
        overlap = alignment.get_overlap(feature.start, feature.end + 1)

    return overlap

def reference_edges(alignment):
    ''' Returns a dictionary of 
    {
        5: reference position mapped to by the left-most query nt in alignment,
        3: reference position mapped to by the right-most query nt in alignment,
    }

    TODO: For consistency with other conventions, these keys should probably
    be 'left' and 'right' instead of 5 and 3.
    '''

    if alignment is None or alignment.is_unmapped:
        return {5: None, 3: None}

    strand = get_strand(alignment)

    if strand == '+':
        edges = {
            5: alignment.reference_start,
            3: alignment.reference_end - 1,
        }
    elif strand == '-':
        edges = {
            5: alignment.reference_end - 1,
            3: alignment.reference_start,
        }

    return edges

def reference_interval(alignment):
    return interval.Interval(alignment.reference_start, alignment.reference_end - 1)

def aligned_tuples(alignment, ref_seq=None):
    tuples = []

    if ref_seq is None:
        aligned_pairs = alignment.get_aligned_pairs(with_seq=True)

        # Remove soft-clipping
        min_i = min(i for i, (q, _, ref_b) in enumerate(aligned_pairs) if ref_b is not None)
        max_i = max(i for i, (q, _, ref_b) in enumerate(aligned_pairs) if ref_b is not None)
        aligned_pairs = aligned_pairs[min_i:max_i + 1]

        for read_i, ref_i, ref_b in aligned_pairs:
            if read_i is None:
                true_read_i = None
                read_b = '-'
                qual = -1
            else:
                true_read_i = true_query_position(read_i, alignment)
                read_b = alignment.query_sequence[read_i].upper()
                qual = alignment.query_qualities[read_i]
            
            if ref_i is not None and ref_b is None:
                # don't include spliced reference positions. (Is this an appropriate check for this situation?)
                continue

            if ref_i is None:
                ref_b = '-'
            else:
                ref_b = ref_b.upper()

            tuples.append((true_read_i, read_b, ref_i, ref_b, qual))

    else:
        aligned_pairs = aligned_pairs_exclude_soft_clipping(alignment)
        
        for read_i, ref_i in aligned_pairs:
            if read_i is None:
                read_b = '-'
                true_read_i = None
                qual = -1
            else:
                true_read_i = true_query_position(read_i, alignment)
                read_b = alignment.query_sequence[read_i].upper()
                qual = alignment.query_qualities[read_i]
                
            if ref_i is None:
                ref_b = '-'
            else:
                ref_b = ref_seq[ref_i].upper()
                
            tuples.append((true_read_i, read_b, ref_i, ref_b, qual))

    return tuples

def total_edit_distance(alignment, ref_seq=None):
    return edit_distance_in_query_interval(alignment, ref_seq=ref_seq)

def edit_distance_in_query_interval(alignment, query_interval=None, ref_seq=None, only_Q30=False):
    if query_interval is None:
        query_interval = interval.Interval(0, np.inf)

    if query_interval.is_empty or alignment is None or alignment.is_unmapped:
        return 0

    distance = 0

    start = query_interval.start
    end = query_interval.end

    tuples = aligned_tuples(alignment, ref_seq)
    if alignment.is_reverse:
        tuples = tuples[::-1]

    first_i = min(i for i, (q, _, _, _, _) in enumerate(tuples) if q is not None and q >= start)
    last_i = max(i for i, (q, _, _, _, _) in enumerate(tuples) if q is not None and q <= end)
    
    for q, q_base, r, r_base, qual in tuples[first_i:last_i + 1]:
        if q_base != r_base:
            if qual >= 30 or not only_Q30:
                distance += 1

    return distance

def get_header(bam_fn):
    saved_verbosity = pysam.set_verbosity(0)
    with pysam.AlignmentFile(bam_fn) as bam_file:
        header = bam_file.header
    pysam.set_verbosity(saved_verbosity)
    return header

def flip_alignment(alignment):
    # Changing the is_reverse flag is functionally equivalent to the read having been
    # reverse complemented before alignment.
    flipped_alignment = copy.deepcopy(alignment)
    flipped_alignment.is_reverse = not alignment.is_reverse
    return flipped_alignment

def fingerprint(al):
    if al is None:
        return None
    else:
        return tuple(al.cigar), al.reference_start, al.reference_name, al.is_reverse

def make_nonredundant(alignments):
    ''' Two alignments of the same read are redundant if they pair the same read bases with the same
    reference bases. Given alignments of the same read, return alignments in which only one representative
    of each equivalent class of redundancy is retained.
    ''' 
    
    nonredundant = {fingerprint(al): al for al in alignments if al is not None}
    
    return list(nonredundant.values())

def make_noncontained(alignments, max_length=np.inf, alignments_contained_in=None):
    ''' Alignment A is contained in alignment B if
    the query interval covered by A is a subset of the query interval covered by B.
    Does not remove als longer than max_length.
    ''' 
    query_intervals = [interval.get_covered(al) for al in alignments]

    if alignments_contained_in is None:
        alignments_contained_in = alignments
        query_intervals_contained_in = query_intervals
    else:
        query_intervals_contained_in = [interval.get_covered(al) for al in alignments_contained_in]

    to_remove = set()
    for first_i in range(len(alignments)):
        for second_i in range(len(alignments_contained_in)):

            if alignments[first_i] == alignments_contained_in[second_i]:
                continue

            if query_intervals[first_i] in query_intervals_contained_in[second_i]:
                if query_intervals[first_i].total_length <= max_length:
                    to_remove.add(first_i)
    
    noncontained = [al for i, al in enumerate(alignments) if i not in to_remove and al is not None]
    
    return noncontained


import numpy as np
from . import utilities
from . import sw
from .adapters_cython import *

primers = {
    'truseq': {
        'R1':         'TCTTTCCCTACACGACGCTCTTCCGATCT',
        'R2':    'GTGACTGGAGTTCAGACGTGTGCTCTTCCGATCT',
    },
    'PE': {
        'R1':         'TCTTTCCCTACACGACGCTCTTCCGATCT', # Note: same as truseq R1
        'R2': 'CGGTCTCGGCATTCCTGCTGAACCGCTCTTCCGATCT',
    },
    'nextera': {
        'R1':     'TCGTCGGCAGCGTCAGATGTGTATAAGAGACAG',
        'R2':    'GTCTCGTGGGCTCGGAGATGTGTATAAGAGACAG',
    },
}

primers['mix_and_match'] = {
    'R1': primers['nextera']['R1'],
    'R2': primers['truseq']['R2'],
}

flow_cell = {
    'P5': 'AATGATACGGCGACCACCGAGATCTACAC',
    'P7': 'CAAGCAGAAGACGGCATACGAGAT',
}

A_tail = 'A' * 10

# For backwards compatibility
truseq_R1_rc = 'AGATCGGAAGAGCGTCGTGTAGGGAAAGA'
truseq_R2_rc = 'AGATCGGAAGAGCACACGTCTGAACTCCAGTCAC'
P5_rc = 'TCTCGGTGGTCGCCGTATCATT'
P7_rc = 'ATCTCGTATGCCGTCTTCTGCTTG'

def build_before_adapters(I7_sequence='', primer_type='truseq', just_primers=False, R1_index='', R2_index=''):
    if just_primers:
        before_R1 = primers[primer_type]['R1'] + R1_index
        before_R2 = primers[primer_type]['R2'] + R2_index
    else:
        before_R1 = flow_cell['P5'] + primers[primer_type]['R1'] + R1_index
        before_R2 = flow_cell['P7'] + utilities.reverse_complement(I7_sequence) + primers[primer_type]['R2'] + R2_index

    return before_R1, before_R2

def build_adapters(**kwargs):
    before_R1, before_R2 = build_before_adapters(**kwargs)

    adapter_in_R1 = utilities.reverse_complement(before_R2) + A_tail
    adapter_in_R2 = utilities.reverse_complement(before_R1) + A_tail

    return adapter_in_R1, adapter_in_R2

def build_adapter_ranges(index_sequence, primer_type='truseq'):
    def make_ranges(construct, names):
        cumulative_lengths = list(np.cumsum(map(len, construct)))
        bounds = zip([0] + cumulative_lengths, cumulative_lengths)
        ranges = zip(names, bounds)
        return ranges
    
    primer_in_R1 = utilities.reverse_complement(primers[primer_type]['R2'])
    primer_in_R2 = utilities.reverse_complement(primers[primer_type]['R1'])

    R1_construct = [primer_in_R1,
                    index_sequence,
                    P7_rc,
                    A_tail,
                   ]
    R1_names = ['R2 primer',
                'I7',
                'P7',
                'A tail',
               ]

    chemistry_only_cycles = 7
    I5_length = 8
    R2_construct = [primer_in_R2[:-(I5_length + chemistry_only_cycles)],
                    primer_in_R2[-(I5_length + chemistry_only_cycles):-(chemistry_only_cycles)],
                    primer_in_R2[-chemistry_only_cycles:],
                    P5_rc,
                    A_tail,
                   ]
    R2_names = ['R1 primer',
                'I5',
                'Chemistry',
                'P7',
                'A tail',
               ]
    
    R1_ranges = make_ranges(R1_construct, R1_names)
    R2_ranges = make_ranges(R2_construct, R2_names)
    return R1_ranges, R2_ranges

def consistent_paired_position(R1_seq,
                               R2_seq,
                               adapter_in_R1,
                               adapter_in_R2,
                               min_comparison_length,
                               max_distance,
                               allow_prefix=True,
                              ):
    R1_positions = find_adapter_positions(R1_seq, adapter_in_R1, min_comparison_length, max_distance)
    R2_positions = find_adapter_positions(R2_seq, adapter_in_R2, min_comparison_length, max_distance)

    if allow_prefix:
        R1_prefix_position = find_adapter(adapter_in_R1, max_distance, R1_seq)
        if R1_prefix_position != len(R1_seq):
            R1_positions.append(R1_prefix_position)
        
        R2_prefix_position = find_adapter(adapter_in_R2, max_distance, R2_seq)
        if R2_prefix_position != len(R2_seq):
            R2_positions.append(R2_prefix_position)

    R1_positions = set(R1_positions)
    R2_positions = set(R2_positions)
    common_positions = R1_positions & R2_positions
    if common_positions:
        return min(common_positions)
    else:
        return None

def trim_by_local_alignment(adapter, seq):
    ''' Try to find a near-exact match. If this fails, do a local alignment. '''
    adapter_prefix_length = 15
    adapter_prefix_bytes = adapter[:adapter_prefix_length].encode()
    seq_bytes = seq.encode()
    trim_at = find_adapter(adapter_prefix_bytes, 1, seq_bytes)

    if trim_at > len(seq) - adapter_prefix_length:
        alignments = sw.generate_alignments(adapter,
                                            seq,
                                            'unpaired_adapter',
                                            max_alignments=1,
                                           )
        if len(alignments) == 1:
            alignment = alignments[0]
            score_diff = 2 * len(alignment['path']) - alignment['score']
            adapter_start_in_seq = sw.first_target_index(alignment['path'])

            if alignment['path'] and score_diff <= 0.5 * len(alignment['path']):
                trim_at = adapter_start_in_seq

    return trim_at



import heapq
import tempfile
import os

def merge(input_files):
    ''' Given input_files, all sorted, returns a generator of the merged
        lines.
    '''
    return heapq.merge(*input_files)

def _sort_chunk(chunk, chunk_file_names):
    ''' Sorts the lines in chunk, writes the sorted lines to a temporary
        file, and appends the temporary file name to chunk_file_names.
    '''
    chunk.sort()
    
    with tempfile.NamedTemporaryFile(mode='w', delete=False) as chunk_file:
        chunk_file_names.append(chunk_file.name)
        for line in chunk:
            chunk_file.write(line)

def external_sort(input_file, sorted_file, chunk_size=3e6):
    ''' Writes the lines in file_handle in sorted order to sorted_file.
        Never loads more than chunk_size lines into memory at the same time.
    '''
    chunk_file_names = []
    
    chunk = []
    for line in input_file:
        chunk.append(line)

        if len(chunk) == chunk_size:
            # When memory is "full", sort and dump to a file.
            _sort_chunk(chunk, chunk_file_names)
            chunk = []
    
    # Sort and dump the last partial fill-up of memory.
    if len(chunk) > 0:
        _sort_chunk(chunk, chunk_file_names)

    chunk_files = [open(fn) for fn in chunk_file_names]
    for line in merge(chunk_files):
        sorted_file.write(line)

    for chunk_file in chunk_files:
        chunk_file.close()

    for chunk_file_name in chunk_file_names:
        os.remove(chunk_file_name)

def sort_simple(file_handle, sorted_file_name):
    ''' Writes the lines in file_handle in sorted order to sorted_file_name.
    '''
    lines = file_handle.readlines()
    lines.sort()
    with open(sorted_file_name, 'w') as sorted_fh:
        for line in lines:
            sorted_fh.write(line)



import unittest
import pysam
import ..sam as sam

MATCH = sam.BAM_CMATCH
DEL = sam.BAM_CDEL
INS = sam.BAM_CINS
SOFT_CLIP = sam.BAM_CSOFT_CLIP

class TestCropAlToRefInt(unittest.TestCase):

    def test_simple(self):
        al = pysam.AlignedSegment()
        al.reference_start = 100
        al.cigar = [(MATCH, 100)]
        al.seq = 'A' * 100

        cropped = pysam.AlignedSegment()
        cropped.reference_start = 150
        cropped.cigar = [(SOFT_CLIP, 50), (MATCH, 10), (SOFT_CLIP, 40)]
        cropped.seq = 'A' * 100

        output = sam.crop_al_to_ref_int(al, 150, 159)
        self.assertEqual(output, cropped)

        output = sam.crop_al_to_ref_int(al, 150, 158)
        self.assertNotEqual(output, cropped)

    def test_ends_in_deletion(self):
        al = pysam.AlignedSegment()
        al.reference_start = 100
        al.cigar = [(MATCH, 50), (DEL, 10), (MATCH, 50)]
        al.seq = 'A' * 100
        
        cropped = pysam.AlignedSegment()
        cropped.reference_start = 100
        cropped.cigar = [(MATCH, 50), (SOFT_CLIP, 50)]
        cropped.seq = 'A' * 100
        
        output = sam.crop_al_to_ref_int(al, 100, 155)
        self.assertEqual(output, cropped)
        
        output = sam.crop_al_to_ref_int(al, 100, 160)
        self.assertNotEqual(output, cropped)
    
    def test_ends_just_after_deletion(self):
        al = pysam.AlignedSegment()
        al.reference_start = 100
        al.cigar = [(MATCH, 50), (DEL, 10), (MATCH, 50)]
        al.seq = 'A' * 100
        
        cropped = pysam.AlignedSegment()
        cropped.reference_start = 100
        cropped.cigar = [(MATCH, 50), (DEL, 10), (MATCH, 1), (SOFT_CLIP, 49)]
        cropped.seq = 'A' * 100
        
        output = sam.crop_al_to_ref_int(al, 100, 160)
        self.assertEqual(output, cropped)
        
        output = sam.crop_al_to_ref_int(al, 100, 155)
        self.assertNotEqual(output, cropped)
    
    def test_starts_in_deletion(self):
        al = pysam.AlignedSegment()
        al.reference_start = 100
        al.cigar = [(MATCH, 50), (DEL, 10), (MATCH, 50)]
        al.seq = 'A' * 100
        
        cropped = pysam.AlignedSegment()
        cropped.reference_start = 160
        cropped.cigar = [(SOFT_CLIP, 50), (MATCH, 50)]
        cropped.seq = 'A' * 100
        
        output = sam.crop_al_to_ref_int(al, 155, 210)
        self.assertEqual(output, cropped)
        
        output = sam.crop_al_to_ref_int(al, 149, 210)
        self.assertNotEqual(output, cropped)
    
    def test_starts_just_before_deletion(self):
        al = pysam.AlignedSegment()
        al.reference_start = 100
        al.cigar = [(MATCH, 50), (DEL, 10), (MATCH, 50)]
        al.seq = 'A' * 100
        
        cropped = pysam.AlignedSegment()
        cropped.reference_start = 149
        cropped.cigar = [(SOFT_CLIP, 49), (MATCH, 1), (DEL, 10), (MATCH, 50)]
        cropped.seq = 'A' * 100
        
        output = sam.crop_al_to_ref_int(al, 149, 210)
        self.assertEqual(output, cropped)
        
        output = sam.crop_al_to_ref_int(al, 150, 210)
        self.assertNotEqual(output, cropped)
    
    def test_ends_just_before_insertion(self):
        al = pysam.AlignedSegment()
        al.reference_start = 100
        al.cigar = [(MATCH, 50), (INS, 10), (MATCH, 50)]
        al.seq = 'A' * 110
        
        cropped = pysam.AlignedSegment()
        cropped.reference_start = 100
        cropped.cigar = [(MATCH, 50), (SOFT_CLIP, 60)]
        cropped.seq = 'A' * 110
        
        output = sam.crop_al_to_ref_int(al, 100, 149)
        self.assertEqual(output, cropped)
        
        output = sam.crop_al_to_ref_int(al, 100, 150)
        self.assertNotEqual(output, cropped)

    def test_ends_just_after_insertion(self):
        al = pysam.AlignedSegment()
        al.reference_start = 100
        al.cigar = [(MATCH, 50), (INS, 10), (MATCH, 50)]
        al.seq = 'A' * 110
        
        cropped = pysam.AlignedSegment()
        cropped.reference_start = 100
        cropped.cigar = [(MATCH, 50), (INS, 10), (MATCH, 1), (SOFT_CLIP, 49)]
        cropped.seq = 'A' * 110
        
        output = sam.crop_al_to_ref_int(al, 100, 150)
        self.assertEqual(output, cropped)
        
        output = sam.crop_al_to_ref_int(al, 100, 149)
        self.assertNotEqual(output, cropped)
    
    def test_starts_just_before_insertion(self):
        al = pysam.AlignedSegment()
        al.reference_start = 100
        al.cigar = [(MATCH, 50), (INS, 10), (MATCH, 50)]
        al.seq = 'A' * 110
        
        cropped = pysam.AlignedSegment()
        cropped.reference_start = 149
        cropped.cigar = [(SOFT_CLIP, 49), (MATCH, 1), (INS, 10), (MATCH, 50)]
        cropped.seq = 'A' * 110
        
        output = sam.crop_al_to_ref_int(al, 149, 210)
        self.assertEqual(output, cropped)
        
        output = sam.crop_al_to_ref_int(al, 150, 210)
        self.assertNotEqual(output, cropped)
    
    def test_starts_just_after_insertion(self):
        al = pysam.AlignedSegment()
        al.reference_start = 100
        al.cigar = [(MATCH, 50), (INS, 10), (MATCH, 50)]
        al.seq = 'A' * 110
        
        cropped = pysam.AlignedSegment()
        cropped.reference_start = 150
        cropped.cigar = [(SOFT_CLIP, 60), (MATCH, 50)]
        cropped.seq = 'A' * 110
        
        output = sam.crop_al_to_ref_int(al, 150, 210)
        self.assertEqual(output, cropped)
        
        output = sam.crop_al_to_ref_int(al, 149, 210)
        self.assertNotEqual(output, cropped)

if __name__ == '__main__':
    unittest.main()



# cython: language_level=3

cpdef int adapter_hamming_distance(char *seq,
                                   char *adapter,
                                   int seq_length,
                                   int adapter_length,
                                   int start,
                                 ):
    ''' Returns the hamming distance between the overlap of seq[start:] and
        adapter.
    '''
    cdef int compare_length = min(adapter_length, seq_length - start)
    cdef int mismatches = 0
    cdef int i

    for i in range(compare_length):
        if seq[start + i] != adapter[i]:
            mismatches += 1

    return mismatches

cpdef simple_hamming_distance(char *first_seq, char *second_seq):
    return adapter_hamming_distance(first_seq, second_seq, len(first_seq), len(second_seq), 0)

def find_adapter(char* adapter, int max_distance, char *seq):
    ''' Returns the leftmost position in seq for which either:
            - seq[position:position + len(adapter)] is within hamming distance
              max_distance of adapter
            - seq[position:] is at least 10 bases long and is within hamming
              distance one of a prefix of adapter
            - seq[position:] exactly matches a prefix of adapter.
    '''
    cdef int seq_length = len(seq)
    cdef int adapter_length = len(adapter)
    cdef int distance, start
    cdef int max_long_prefix_distance = min(max_distance, 1)
        
    for start in range(seq_length - adapter_length + 1):
        distance = adapter_hamming_distance(seq,
                                            adapter,
                                            seq_length,
                                            adapter_length,
                                            start,
                                           )
        if distance <= max_distance:
            return start
    
    for start in range(seq_length - adapter_length + 1, seq_length):
        distance = adapter_hamming_distance(seq,
                                            adapter,
                                            seq_length,
                                            adapter_length,
                                            start,
                                           )
        if distance == 0:
            return start
        elif seq_length - start >= 10 and distance <= max_long_prefix_distance:
            return start
    
    # Convention: position of seq_length means no position was found
    return seq_length

def find_adapter_positions(read, adapter, int min_comparison_length, int max_distance):
    ''' Temporary for backwards compatibility. '''
    cdef int read_length = len(read)
    cdef int adapter_length = len(adapter)
    cdef int max_start = len(read) - min_comparison_length
    cdef int distance, start
        
    positions = [] 
    for start in range(max_start + 1):
        distance = adapter_hamming_distance(read,
                                            adapter,
                                            read_length,
                                            adapter_length,
                                            start,
                                           )
        if distance <= max_distance:
            positions.append(start)
    return positions



import colorsys
import io
import functools
from collections import Counter

import scipy.stats
import seaborn as sns
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import PIL

from .. import utilities
from . import define_igv_colors

igv_colors = define_igv_colors.normalized_rgbs

blues_cdict = {
    'red':   ((0.0, 1.0, 1.0),
              (1.0, 0.0, 0.0)),
    'green': ((0.0, 1.0, 1.0),
              (1.0, 0.0, 0.0)),
    'blue':  ((0.0, 1.0, 1.0),
              (1.0, 1.0, 1.0)),
}

blues = matplotlib.colors.LinearSegmentedColormap('blues', blues_cdict, 10000)
blues.set_over('black')

reds_cdict = {
    'red':   ((0.0, 1.0, 1.0),
              (1.0, 1.0, 1.0)),
    'green': ((0.0, 1.0, 1.0),
              (1.0, 0.0, 0.0)),
    'blue':  ((0.0, 1.0, 1.0),
              (1.0, 0.0, 0.0)),
}

reds = matplotlib.colors.LinearSegmentedColormap('reds', reds_cdict, 10000)
reds.set_over('black')

greens_cdict = {
    'red':   ((0.0, 1.0, 1.0),
              (1.0, 0.0, 0.0)),
    'green': ((0.0, 1.0, 1.0),
              (1.0, 1.0, 1.0)),
    'blue':  ((0.0, 1.0, 1.0),
              (1.0, 0.0, 0.0)),
}

greens = matplotlib.colors.LinearSegmentedColormap('greens', greens_cdict, 10000)

def optional_ax(original_function):
    @functools.wraps(original_function)
    def possibly_new_ax(*args, **kwargs):
        ax_given = kwargs.get('ax')
        if not ax_given:
            fig, ax = plt.subplots(figsize=(12, 8))
            kwargs['ax'] = ax
        
        result = original_function(*args, **kwargs)
        
        figure_file_name = kwargs.get('save_as')
        if figure_file_name:
            assert not ax_given, 'Can\'t give ax and figure_file_name'
            fig.savefig(figure_file_name, bbox_inches='tight')
            plt.close(fig)
        
        return result
    
    return possibly_new_ax

def add_commas_to_ticks(ax, which='y'):
    def commas_formatter(x, pos):
        return f'{int(x):,}'

    tick_formatter = matplotlib.ticker.FuncFormatter(commas_formatter)
    if which == 'y':
        axis = ax.yaxis
    elif which == 'x':
        axis = ax.xaxis
    else:
        raise ValueError(which)

    axis.set_major_formatter(tick_formatter)

@optional_ax
def enhanced_scatter(xs, ys,
                     ax=None,
                     data=None,
                     colors=None,
                     do_fit=False,
                     show_correlation=False,
                     show_p_value=False,
                     show_beta=True,
                     hists_location=None,
                     hist_bins=100,
                     hist_range=None,
                     hist_alpha=0.2,
                     hist_size_ratio=0.1,
                     hist_colors=None,
                     hist_share_max=True,
                     marker_size=4,
                     text_size=14,
                     text_weight='normal',
                     text_location='lower right',
                     color_by_correlation=False,
                     fit_line_kwargs={'color': 'black',
                                      'alpha': 0.3,
                                     },
                     r_digits=2,
                     variance=False,
                     in_log_space=False,
                     big_fit_line=False,
                     colorbar=False,
                     remove_y_hist=False,
                     remove_x_hist=False,
                     label=None,
                     alpha=None,
                     **scatter_kwargs,
                    ):

    if data is not None:
        x_label = xs
        y_label = ys
        color_label = colors
        xs = data[xs]
        ys = data[ys]
        try:
            if colors in data:
                colors = data[colors]
        except TypeError:
            pass
    else:
        x_label = ''
        y_label = ''
        color_label = ''

    xs = np.asarray(xs)
    ys = np.asarray(ys)

    same_lists = np.allclose(xs, ys)

    if isinstance(colors, str) and colors == 'density':
        if not same_lists and len(xs) > 2:
            indices = np.arange(len(xs))
            np.random.shuffle(indices)
            random_indices = indices[:5000]

            sampled_points = np.vstack([xs[random_indices], ys[random_indices]])
            points = np.vstack([xs, ys])
            kernel = scipy.stats.gaussian_kde(sampled_points)
            colors = kernel(points)
        else:
            colors = 'black'
    elif colors is not None:
        if isinstance(colors, str):
            # colors is a single color name
            pass
        else:
            #colors = np.asarray(colors)
            pass
    
    if same_lists:
        do_fit = False

    kwargs = {
        's': marker_size,
        'linewidths' : (0,),
        'label': label,
        'alpha': alpha,
        **scatter_kwargs,
    }

    scatter = ax.scatter(xs, ys, c=colors, **kwargs)
    ax.set_xlabel(x_label, size=16)
    ax.set_ylabel(y_label, size=16)
    fig = ax.figure
    ax_position = ax.get_position()

    cax = None

    if not isinstance(colors, str):
        if colorbar:
            cax = fig.add_axes((ax_position.x0 + ax_position.width * 1.17,
                                ax_position.y1 - ax_position.height * 0.4,
                                ax_position.width * 0.03,
                                ax_position.height * 0.35,
                               ),
                              )
            fig.colorbar(scatter, cax=cax)
            cax.set_title(color_label)

    if isinstance(text_location, tuple):
        x, y = text_location
        horizontal_alignment = 'center'
        vertical_alignment = 'top'
        x_sign = 1
        y_sign = -1

    if 'left' in text_location:
        x = 0
        x_sign = 1
        horizontal_alignment = 'left'
    elif 'right' in text_location:
        x = 1
        x_sign = -1
        horizontal_alignment = 'right'

    if 'upper' in text_location:
        y = 1
        y_sign = -1
        vertical_alignment = 'top'
    elif 'lower' in text_location:
        y = 0
        y_sign = 1
        vertical_alignment = 'bottom'

    if text_location == 'above':
        x = 0.5
        y = 1
        x_sign = 1
        y_sign = 1.05
        vertical_alignment = 'bottom'
        horizontal_alignment = 'center'
        x_offset = 0
        y_offset = 0
    else:
        x_offset = 10
        y_offset = 0.5 * text_size

    text_kwargs = {
        'xy': (x, y),
        'xycoords': 'axes fraction',
        'textcoords': 'offset points',
        'horizontalalignment': horizontal_alignment,
        'verticalalignment': vertical_alignment,
        'fontsize': text_size,
        'family': 'serif',
        'weight': text_weight,
    }
    
    if do_fit:
        fit = np.polyfit(xs, ys, 1)
        beta, constant = fit
        fit_function = np.poly1d(fit)
        x_lims = ax.get_xlim()
        if big_fit_line:
            plot_lims = (min(xs) - (max(xs) - min(xs)), max(xs) + (max(xs) - min(xs)))
        else:
            plot_lims = x_lims
        ax.plot(plot_lims, fit_function(plot_lims), **fit_line_kwargs)
        ax.set_xlim(*x_lims)
        
        if show_beta == 'fit':
            ax.annotate(r'$y = {0:0.2f} x {2:s} {1:0.2f}$'.format(beta, abs(constant), '+' if constant > 0 else '-'),
                        xytext=(x_sign * x_offset, y_sign * y_offset * 4),
                        **text_kwargs)
        elif show_beta:
            ax.annotate(r'$\beta$ = {:0.2f}'.format(beta),
                        xytext=(x_sign * x_offset, y_sign * y_offset * 4),
                        **text_kwargs)

    if in_log_space:
        # Values have been log'ed before being passed in, so exponentiate them
        # to recover correct correlation.
        r, p = scipy.stats.pearsonr(in_log_space**xs, in_log_space**ys)
    else:
        r, p = scipy.stats.pearsonr(xs, ys)

    if variance:
        r_part = '$r^2$' + '= {:0.{digits}f}'.format(r**2, digits=r_digits)
    else:
        r_part = 'r = {:0.{digits}f}'.format(r, digits=r_digits)
    if show_p_value:
        p_part = ', p={:0.2e}'.format(p)
    else:
        p_part = ''

    text = r_part + p_part

    if color_by_correlation:
        text_kwargs['color'] = matplotlib.cm.seismic(0.5 * r + 0.5)

    if show_correlation:
        ax.annotate(text,
                    xytext=(x_sign * x_offset, y_sign * y_offset),
                    **text_kwargs,
        )

    max_x = None
    max_y = None

    if hists_location is not None:
        if hists_location == 'inside':
            bottom = ax_position.y0
            left = ax_position.x0
        elif hists_location == 'outside':
            bottom = ax_position.y1
            left = ax_position.x1
        
        common_kwargs = {
            'alpha': hist_alpha,
            'histtype': 'stepfilled',
        }

        if isinstance(hist_bins, int):
            hist_bins = {which: hist_bins for which in ['x', 'y']} 

        if hist_colors is None:
            hist_colors = {which: colors if isinstance(colors, str) else 'black' for which in ['x', 'y']}

        if hist_range is None:
            hist_range = {
                'x': (min(xs), max(xs)),
                'y': (min(ys), max(ys)),
            }

        ax_x = fig.add_axes((ax_position.x0, bottom, ax_position.width, ax_position.height * hist_size_ratio), sharex=ax)
        n_x, *rest = ax_x.hist(xs, range=hist_range['x'], bins=hist_bins['x'], color=hist_colors['x'], **common_kwargs)
        ax_x.axis('off')

        ax_y = fig.add_axes((left, ax_position.y0, ax_position.width * hist_size_ratio, ax_position.height), sharey=ax)
        n_y, *rest = ax_y.hist(ys, range=hist_range['y'], orientation='horizontal', bins=hist_bins['y'], color=hist_colors['y'], **common_kwargs)
        ax_y.axis('off')

        if hist_share_max:
            max_n = max(max(n_x), max(n_y))
            max_x = max_n
            max_y = max_n
        else:
            max_x = max(n_x)
            max_y = max(n_y)

        ax_x.set_ylim(0, max_x * 1.1)
        ax_y.set_xlim(0, max_y * 1.1)

        if remove_x_hist:
            fig.delaxes(ax_x)
        if remove_y_hist:
            fig.delaxes(ax_y)
    else:
        ax_x = None
        ax_y = None

    return fig, {'scatter': ax, 'hist_x': ax_x, 'hist_y': ax_y, 'colorbar': cax, 'hist_maxes': {'x': max_x, 'y': max_y}}

def draw_diagonal(ax, anti=False, color='black', **kwargs):
    if anti:
        xs, ys = [0, 1], [1, 0]
    else:
        xs, ys = [0, 1], [0, 1]
        if ax.get_xlim() != ax.get_ylim():
            print('warning: diagonal in non-equal axes')
    
    ax.plot(xs, ys,
            transform=ax.transAxes,
            color=color,
            **kwargs)

def draw_zeros_and_diagonal(ax, color='black', alpha=0.5):
    draw_diagonal(ax, color=color, alpha=alpha)
    ax.axhline(0, color=color, alpha=alpha)
    ax.axvline(0, color=color, alpha=alpha)

def label_scatter_plot(ax, xs, ys, labels,
                       data=None,
                       color=None,
                       to_label=slice(None),
                       vector='orthogonal',
                       initial_distance=50,
                       distance_increment=10,
                       arrow_alpha=0.2,
                       manual_ratios=None,
                       manual_alignments=None,
                       text_kwargs={'size': 10},
                       avoid=True,
                       avoid_axis_labels=False,
                       avoid_existing=False,
                       min_arrow_distance=10,
                       slope=1,
                      ):
    if data is not None:
        xs = data[xs]
        ys = data[ys]
        if color is not None:
            if color in data:
                color = data[color]
            else:
                color = [color]*len(xs)
        if labels in data:
            labels = data[labels]
        elif labels == data.index.name:
            labels = data.index
        else:
            raise IndexError

    def attempt_text(x, y, site, distance, vector, color):
        if vector == 'orthogonal':
            x_offset = np.sign(x * slope - y) * distance
            y_offset = -np.sign(x * slope - y) * distance
            ha = 'center'
            va = 'top' if y_offset < 0 else 'bottom'
        elif vector == 'upper left':
            x_offset = -distance
            y_offset = distance
            ha = 'center'
            va = 'bottom'
        elif vector == 'lower left':
            x_offset = -distance
            y_offset = -distance
            ha = 'center'
            va = 'top'
        elif vector == 'lower right':
            x_offset = distance
            y_offset = -distance
            ha = 'center'
            va = 'top'
        elif vector == 'upper right':
            x_offset = distance
            y_offset = distance
            ha = 'center'
            va = 'bottom'
        elif vector == 'radial':
            norm = np.linalg.norm([x, y])
            x_offset = x * distance / norm
            y_offset = y * distance / norm
            ha = 'center'
            va = 'top' if y_offset < 0 else 'bottom'
        elif vector == 'above':
            x_offset = 0
            y_offset = distance
            ha = 'center'
            va = 'bottom'
        elif vector == 'below':
            x_offset = 0
            y_offset = -distance
            ha = 'center'
            va = 'top'
        elif vector == 'sideways':
            x_offset = distance
            y_offset = 0
            ha, va = 'left', 'center'
        elif vector == 'right':
            x_offset = distance
            y_offset = 0
            ha, va = 'left', 'center'
        elif vector == 'left':
            x_offset = -distance
            y_offset = 0
            ha, va = 'right', 'center'
        elif vector == 'manual':
            x_ratio, y_ratio = manual_ratios
            ha, va = manual_alignments
            x_offset = distance * x_ratio
            y_offset = distance * y_ratio
        else:
            raise ValueError(vector)

        text = ax.annotate(site,
                           xy=(x, y),
                           xycoords=('data', 'data'),
                           xytext=(x_offset, y_offset),
                           textcoords='offset points',
                           ha=ha,
                           va=va,
                           color=color,
                           **text_kwargs,
                          )

        if avoid:
            ax.figure.canvas.draw()

        return text, text.get_window_extent(), (x, y, x_offset, y_offset)

    ax.figure.canvas.draw()

    starting_labels = []

    if avoid_axis_labels:
        to_add = [ax.xaxis.get_label(), ax.yaxis.get_label()] + ax.get_yticklabels() + ax.get_xticklabels()
        starting_labels.extend(to_add)

    if avoid_existing:
        to_add = [c for c in ax.get_children() if isinstance(c, matplotlib.text.Annotation)]
        starting_labels.extend(to_add)

    bboxes = []
    for label in starting_labels:
        try:
            bboxes.append(label.get_window_extent())
        except:
            pass

    if isinstance(vector, str):
        vector = np.array([vector]*len(xs))

    if color is None:
        color = 'black'

    if isinstance(color, str):
        color = np.array([color]*len(xs))

    tuples = zip(
        xs[to_label],
        ys[to_label],
        labels[to_label],
        vector[to_label],
        color[to_label],
    )

    for x, y, label, vec, color in tuples:
        distance = initial_distance
        text, bbox, coords = attempt_text(x, y, label, distance, vec, color)

        while avoid and any(bbox.overlaps(other_bbox) for other_bbox in bboxes):
            text.remove()
            distance += distance_increment
            text, bbox, coords = attempt_text(x, y, label, distance, vec, color)
            if distance >= distance_increment * 50:
                print(f'gave up on {label}')
                break
        
        if distance >= min_arrow_distance:
            x, y, x_offset, y_offset = coords
            ax.annotate('',
                        xy=(x, y),
                        xycoords=('data', 'data'),
                        xytext=(x_offset, y_offset),
                        textcoords=('offset points', 'offset points'),
                        arrowprops={'arrowstyle': '-',
                                    'alpha': arrow_alpha,
                                    'color': color,
                                   },
                       )

        bboxes.append(bbox)

def evenly_spaced_jet_colors(n):
    jet_colors = [matplotlib.cm.jet(x) for x in np.linspace(0, 1, n)]
    return jet_colors

def assign_colors_by_ranks(key_to_ranks, index=0):
    colors = evenly_spaced_jet_colors(len(key_to_ranks))
    key_to_color = {key: colors[key_to_ranks[key][index]] for key in key_to_ranks}
    return key_to_color

def color_labels(labels, name_to_color):
    for label in labels:
        color = name_to_color.get(label.get_text(), None)
        if color:
            label.set_color(color)

def apply_alpha(color, alpha, multiplicative=False):
    r, g, b, a =  matplotlib.colors.colorConverter.to_rgba(color)
    if multiplicative:
        a *= alpha
    else:
        a = alpha
    return (r, g, b, a)

def scale_darkness(color, scale):
    ''' Increases darkness of color by converting to hsl and dividing l value
    by scale. adapted from https://stackoverflow.com/a/49601444
    '''
    rgb = matplotlib.colors.to_rgb(color)
    h, l, s = colorsys.rgb_to_hls(*rgb)
    scaled_rgb = colorsys.hls_to_rgb(h, max(0, min(1, l / scale)), s)
    scaled_hex = matplotlib.colors.to_hex(scaled_rgb)
    
    return scaled_hex

def force_integer_ticks(axis):
    axis.set_major_locator(matplotlib.ticker.MaxNLocator(interger=True))

@optional_ax
def plot_counts(l, ax=None, log_scales=None, normalize=False, **kwargs):
    if log_scales is None:
        log_scales = set()

    counts = Counter(l)
    ys = utilities.counts_to_array(counts)

    first_nonzero = ys.nonzero()[0][0]
    xs = np.arange(first_nonzero, len(ys))
    ys = ys[first_nonzero:]
    if normalize:
        ys = ys / ys.sum()

    ax.plot(xs, ys, **kwargs)

    if 'x' in log_scales:
        ax.set_xscale('log')
    else:
        ax.set_xlim(-0.01 * len(ys), 1.01 * len(ys))

    if 'y' in log_scales:
        ax.set_yscale('log')
    else:
        ax.set_ylim(0)

def make_stacked_Image(figs, orientation='vertical'):
    ims = []

    for fig in figs:
        with io.BytesIO() as buffer:
            fig.savefig(buffer, format='png', bbox_inches='tight')
            im = PIL.Image.open(buffer)
            im.load()
            ims.append(im)

        plt.close(fig)
        
    if not ims:
        return None

    if orientation == 'vertical':
        total_height = sum(im.height for im in ims)
        max_width = max(im.width for im in ims)

        stacked_im = PIL.Image.new('RGBA', size=(max_width, total_height), color='white')
        y_start = 0
        for im in ims:
            stacked_im.paste(im, (max_width - im.width, y_start))
            y_start += im.height

    else:
        max_height = max(im.height for im in ims)
        total_width = sum(im.width for im in ims)

        stacked_im = PIL.Image.new('RGBA', size=(total_width, max_height), color='white')
        x_start = 0
        for im in ims:
            stacked_im.paste(im, (x_start, max_height - im.height))
            x_start += im.width

    return stacked_im

def assign_categorical_colors(series, palette=None, sort=True):
    if palette is None:
        palette = sns.color_palette()

    values = series.unique()

    if sort:
        values = sorted(values)

    value_to_color = dict(zip(values, palette))
    colors = series.map(value_to_color)

    return colors, value_to_color

def draw_categorical_legend(value_to_color,
                            ax,
                            font_size=12,
                            legend_location='upper left',
                            order=None,
                            manual_xy=None,
                            aliases=None,
                            title=None,
                           ):

    if legend_location == 'upper left':
        xy = (0.04, 0.96)
        va = 'top'
        ha  = 'left'
    elif legend_location == 'upper middle':
        xy = (0.5, 0.96)
        va = 'top'
        ha = 'center'
    elif legend_location == 'middle left':
        xy = (0.04, 0.5)
        va = 'center'
        ha = 'left'
    elif legend_location == 'upper right':
        xy = (0.96, 0.96)
        va = 'top'
        ha = 'right'
    elif legend_location == 'lower right':
        xy = (0.96, 0.4)
        va = 'bottom'
        ha = 'right'
    elif legend_location == 'lower left':
        xy = (0.06, 0.4)
        va = 'bottom'
        ha = 'left'
    elif legend_location == 'middle':
        xy = (0.5, 0.5)
        va = 'center'
        ha = 'center'

    if manual_xy is not None:
        xy = manual_xy

    if order is None:
        order = sorted(value_to_color)

    if aliases is None:
        aliases = {}

    if title != None:
        order = [f'{title}:'] + order

    for i, category in enumerate(order):
        if category == f'{title}:':
            color = 'black'
        else:
            color = value_to_color[category]
        ax.annotate(aliases.get(category, category),
                    xy=xy,
                    xycoords='axes fraction',
                    xytext=(0, -(font_size + 3) * i),
                    textcoords='offset points',
                    color=color,
                    va=va,
                    ha=ha,
                    size=font_size,
                   )


import hashlib
import itertools
import logging

from collections import defaultdict, Counter
from functools import partial

import scipy.optimize
import scipy.stats
import numpy as np
import pysam

from . import sam, fastq, utilities, genomes, paired_end
from .utilities import base_order, base_to_index, base_to_complement_index

change_order = [('A', 'G'),
                ('T', 'C'),
                ('G', 'A'),
                ('C', 'T'),
                ('G', 'C'),
                ('C', 'G'),
                ('A', 'C'),
                ('T', 'G'),
                ('A', 'T'),
                ('T', 'A'),
                ('G', 'T'),
                ('C', 'A'),
                ('A', 'A'),
                ('T', 'T'),
                ('G', 'G'),
                ('C', 'C'),
                ('N', 'N'),
                ('-', '-'),
                ('N', 'A'),
                ('N', 'C'),
                ('N', 'T'),
                ('N', 'G'),
                ('N', '-'),
                ('A', 'N'),
                ('C', 'N'),
                ('T', 'N'),
                ('G', 'N'),
                ('-', 'N'),
                ('-', 'A'),
                ('-', 'C'),
                ('-', 'T'),
                ('-', 'G'),
                ('A', '-'),
                ('C', '-'),
                ('T', '-'),
                ('G', '-'),
               ]

coords_order = [(base_to_index[ref_char], base_to_index[read_char])
                for ref_char, read_char in change_order]

def clopper_pearson(x, n, alpha=0.05):
    if n == 0:
        return 0., 0.
    elif x == 0:
        cdf = lambda t: scipy.stats.binom.cdf(x, n, t) - alpha
        u = scipy.optimize.brentq(cdf, 0, 1)
        l = 0.
        mle = 0.
    elif x == n:
        sf = lambda t: scipy.stats.binom.sf(x - 1, n, t) - alpha
        u = 1.
        l = scipy.optimize.brentq(sf, 0, 1)
        mle = 1.
    else:
        cdf = lambda t: scipy.stats.binom.cdf(x, n, t) - (alpha / 2)
        sf = lambda t: scipy.stats.binom.sf(x - 1, n, t) - (alpha / 2)
        u = scipy.optimize.brentq(cdf, 0, 1)
        l = scipy.optimize.brentq(sf, 0, 1)
        mle = float(x) / n
    return mle - l, u - mle

def get_ineligible_positions(file_name, just_single_positions=False):
    ''' Loads a set of positions at which variants should be ignored (i.e.,
        known SNPs) from file_name. Each line of file_name should consist of
        either
        ref_seq_name
        to indiciate an entire rname that is ineligible, or
        ref_seq_name,position
        to indicate a single ineligible position, or
        ref_seq_name,start-end
        to indicate a range of ineligible positions (i.e. rDNA in yeast).
    '''
    ineligible_positions = set()
    ineligible_rnames = set()
    for line in open(file_name):
        fields = line.strip().split(',')
        if len(fields) == 1:
            ineligible_rnames.add(fields[0])
        elif len(fields) == 2:
            ref_seq_name, position = line.strip().split(',')
            range_bounds = position.split('-')
            if len(range_bounds) == 2:
                if not just_single_positions:
                    start, end = map(int, range_bounds)
                    for p in range(start, end):
                        ineligible_positions.add((ref_seq_name, p))
            else:
                ineligible_positions.add((ref_seq_name, int(position)))
    return ineligible_positions, ineligible_rnames

def hash_ineligible_positions(file_name):
    ''' Produces SHA1 hash of file_name contents. Assembled from various
        stackoverflow answers.
    '''
    sha1 = hashlib.sha1()
    with open(file_name) as fh:
        for chunk in iter(lambda: fh.read(sha1.block_size), ''):
            sha1.update(chunk)
    return sha1.hexdigest()

def count_mismatches(sam_lines,
                     max_read_length,
                     ineligible_positions,
                     ineligible_rnames=[],
                     max_qual=fastq.MAX_QUAL,
                    ):
    ''' max_read_length - a hack to make this have the same argument signature
                          as count_pe_mismatches
        ineligible_positions - set of reference positions to ignore
    '''
    circles_type_shape = (max_read_length / 2 + 1,
                          max_qual + 1,
                          len(base_order),
                          len(base_order),
                         )
    type_counts = np.zeros(shape=circles_type_shape, dtype=int)

    mismatch_ref_positions = []
    
    num_reads_skipped = 0
    total_no_indel_bases = 0
    total_no_SNP_bases = 0

    def process_alignment(sam_line):
        no_SNP_bases = 0
        # Flipping strands is kind of arbitrary for circular products.
        if sam_line['strand'] == '+':
            index_lookup = base_to_index
        else:
            index_lookup = base_to_complement_index

        alignment = sam.produce_alignment(sam_line)
        ref_seq_name = sam_line['RNAME']
        read_length = len(sam_line['SEQ'])
       
        for ref_char, read_char, qual, ref_pos, read_pos in alignment:
            # Skip known SNP positions
            if ref_seq_name in ineligible_rnames or (ref_seq_name, ref_pos) in ineligible_positions:
                continue
            no_SNP_bases += 1

            from_left = read_pos
            from_right = read_length - 1 - read_pos
            from_edge = min(from_left, from_right)
            
            ref_index = index_lookup[ref_char]
            read_index = index_lookup[read_char]
            coords = (from_edge, qual, ref_index, read_index)
            type_counts[coords] += 1
            
            if ref_char != read_char:
                #if qual >= 60:
                #    print sam_line
                #    print ref_char, read_char, qual, read_pos

                close_to_edge = from_edge < 20
                if not close_to_edge and qual > 90:
                    mismatch_ref_position = (ref_seq_name,
                                             ref_pos,
                                             ref_char,
                                             read_char,
                                             1,
                                            )
                    mismatch_ref_positions.append(mismatch_ref_position)

        return no_SNP_bases

    parsed_lines = (sam.parse_line(line) for line in sam_lines)
    
    for parsed_line in parsed_lines:
        skip = False
        if parsed_line['mapped'] == False:
            skip = True
        
        if parsed_line['MAPQ'] == 0:
            skip = True

        if sam.contains_indel(parsed_line) or sam.contains_soft_clipping(parsed_line):
            skip = True
        
        if skip:
            num_reads_skipped += 1
            continue

        total_no_indel_bases += len(parsed_line['SEQ'])
        
        no_SNP_bases = process_alignment(parsed_line)
        total_no_SNP_bases += no_SNP_bases

    mismatch_info = {'type_counts': type_counts,
                     'ref_positions': mismatch_ref_positions,
                     'num_reads_skipped': num_reads_skipped,
                     'no_indel_bases_count': total_no_indel_bases,
                     'no_SNP_bases_count': total_no_SNP_bases,
                    }
    return mismatch_info

def count_pe_mismatches(sam_file,
                        mappings,
                        max_read_length,
                        ineligible_positions,
                        ineligible_rnames=[],
                        max_qual=fastq.MAX_QUAL,
                        threshold_qual=fastq.MAX_QUAL,
                       ):
    pe_type_shape = (2 * max_read_length,
                     max_qual + 1,
                     len(base_order),
                     len(base_order),
                    )
    type_counts = np.zeros(shape=pe_type_shape, dtype=int)
    
    mismatch_ref_positions = []

    num_reads_skipped = 0
    total_no_indel_bases = 0
    total_no_SNP_bases = 0

    mismatches_per_read = Counter()
    mismatch_qnames = Counter()
    indel_qnames = Counter()
    
    def process_alignment(mapping):
        mismatches_this_read = 0
        no_SNP_bases = 0
        
        if mapping.is_reverse:
            strand = '-'
        else:
            strand = '+'

        # R1s are privileged, R2s need to be flipped.
        if mapping.is_read1:
            if strand == '+':
                index_lookup = base_to_index
            else:
                index_lookup = base_to_complement_index
        elif mapping.is_read2:
            if strand == '+':
                index_lookup = base_to_complement_index
            else:
                index_lookup = base_to_index

        alignment = sam.produce_alignment(mapping)
        rname = sam_file.getrname(mapping.tid)
        
        if rname not in ineligible_rnames:
            for ref_char, read_char, qual, ref_pos, read_pos in alignment:
                # Skip known SNP positions
                if (rname, ref_pos) in ineligible_positions:
                    continue
                no_SNP_bases += 1
                
                # Mappings may have overlapping regions truncate. We still want
                # from_left and from_right to reflect distances from the
                # untruncated edges.
                if strand == '+':
                    from_left = read_pos
                    from_right = max_read_length - 1 - read_pos
                    position = from_left
                else:
                    from_left = read_pos - (mapping.query_alignment_end - max_read_length)
                    from_right = mapping.query_alignment_end - 1 - read_pos
                    position = from_right
                
                if mapping.is_read2:
                    position = position + max_read_length
                
                ref_index = index_lookup[ref_char]
                read_index = index_lookup[read_char]
                coords = (position, qual, ref_index, read_index)
                type_counts[coords] += 1
                
                if ref_char != read_char:
                    from_edge = min(from_left, from_right)
                    close_to_edge = from_edge < 10
                    if not close_to_edge and qual >= threshold_qual:
                        mismatch_ref_position = (rname,
                                                 ref_pos,
                                                 ref_char,
                                                 read_char,
                                                 1,
                                                )
                        mismatch_ref_positions.append(mismatch_ref_position)
                        mismatch_qnames[mapping.qname] += 1
                        mismatches_this_read += 1

        return no_SNP_bases, mismatches_this_read

    mapping_pairs = itertools.izip(*[mappings]*2)

    for R1_mapping, R2_mapping in mapping_pairs:
        skip = False

        if R1_mapping.is_unmapped or R2_mapping.is_unmapped:
            skip = True
        
        if R1_mapping.tid != R2_mapping.tid:
            skip = True
        
        # Ignore, for now, reads with indels or clipping
        if sam.contains_indel_pysam(R1_mapping) or sam.contains_indel_pysam(R2_mapping):
            if sam.contains_indel_pysam(R1_mapping):
                if sam.indel_distance_from_edge(R1_mapping.cigar) > 10:
                    indel_qnames[R1_mapping.qname] += 1
            else:
                if sam.indel_distance_from_edge(R2_mapping.cigar) > 10:
                    indel_qnames[R2_mapping.qname] += 1

            skip = True
        if sam.contains_soft_clipping_pysam(R1_mapping) or sam.contains_soft_clipping_pysam(R2_mapping):
            skip = True
        
        if skip:
            num_reads_skipped += 1
            continue
       
        if R1_mapping.seq != None:
            total_no_indel_bases += len(R1_mapping.seq)
        if R2_mapping.seq != None:
            total_no_indel_bases += len(R2_mapping.seq)
        
        R1_no_SNP_bases, R1_mismatches_this_read = process_alignment(R1_mapping)
        total_no_SNP_bases += R1_no_SNP_bases
        R2_no_SNP_bases, R2_mismatches_this_read = process_alignment(R2_mapping)
        total_no_SNP_bases += R2_no_SNP_bases

        mismatches_this_pair = R1_mismatches_this_read + R2_mismatches_this_read
        mismatches_per_read[mismatches_this_pair] += 1
        #if mismatches_this_pair > 1:
        #    print R1_parsed['QNAME'], mismatches_this_pair

    mismatches_per_read = utilities.counts_to_array(mismatches_per_read)
        
    mismatch_info = {'type_counts': type_counts,
                     'ref_positions': mismatch_ref_positions,
                     'num_reads_skipped': num_reads_skipped,
                     'no_indel_bases_count': total_no_indel_bases,
                     'no_SNP_bases_count': total_no_SNP_bases,
                     'mismatches_per_read': mismatches_per_read,
                     'mismatch_qnames': mismatch_qnames,
                     'indel_qnames': indel_qnames,
                    }
    return mismatch_info

def compute_average_qualities(position_type_counts):
    position_qs = position_type_counts.sum(axis=(2, 3))
    position_average_qs = map(utilities.mean_from_histogram, position_qs)
    return position_average_qs

def compute_fractions_miscalled_as(position_type_counts, min_q=0):
    ''' Given an array of type counts whose indices are
    (position in read, quality score, reference identity, read identity),
    computes the fraction of all base calls at each position that were
    misidentified as each identity.
    '''
    num_positions, _, _, _ = position_type_counts.shape
    all_rates = np.zeros((num_positions, 6))
    for p in range(num_positions):
        position_ms = position_type_counts[p][min_q:].sum(axis=0)
        calls_of_each_type = position_ms.sum(axis=0)
        good_calls = position_ms.diagonal()
        bad_calls = calls_of_each_type - good_calls
        total_calls = calls_of_each_type.sum()
        # rates is the fraction of all calls at this position that were of each
        # type when they shouldn't have been
        rates = np.true_divide(bad_calls, max(1, total_calls))
        all_rates[p] = rates[:6]
    return all_rates

def compute_overall_rate(type_counts, min_q=None, max_q=None):
    ''' Return information about overall mismatch rates in type_counts. '''
    if min_q == None:
        min_q = len(type_counts) - 1
    if max_q == None:
        max_q = len(type_counts) - 1

    # Compute expected overall mismatch rate given quality distribution
    q_counts = type_counts[min_q:max_q + 1].sum(axis=1).sum(axis=1)
    es = np.array([10**(-q / 10.) for q in range(min_q, max_q + 1)])
    if q_counts.sum() == 0:
        expected_rate = 0
    else:
        expected_rate = np.dot(q_counts, es) / q_counts.sum()

    # Compute actual overall mismatch rate and error bars
    counts = type_counts[min_q:max_q + 1].sum(axis=0)
    total_bases = int(counts.sum())
    total_matches = counts.diagonal().sum()
    total_mismatches = total_bases - total_matches
    if total_bases != 0:
        overall_rate = float(total_mismatches) / total_bases
    else:
        overall_rate = 0. 
    overall_err = clopper_pearson(total_mismatches, total_bases)
    overall_err = np.array(overall_err).T

    rate_info = {
        'total_bases': total_bases,
        'total_mismatches': total_mismatches,
        'overall_rate': overall_rate,
        'overall_err': overall_err,
        'expected_rate': expected_rate,
    }
    return rate_info

def compute_rates(type_counts, min_q, max_q):
    ''' Computes all categories of mismatch rates from type_counts. '''
    counts = type_counts[min_q:max_q + 1].sum(axis=0)
    # counts is a 6x6 matrix of (ref_char, read_char) counts in the 
    # eligible quality range
    
    mismatch_counts = []
    rates = []
    errs = []
    for ref_coord, read_coord in coords_order[:12]:
        ref_total = sum(counts[ref_coord])
        
        if ref_total != 0:
            rate = float(counts[ref_coord, read_coord]) / ref_total
        else:
            rate = 0.
        rates.append(rate)
        
        mismatch_count = counts[ref_coord, read_coord]
        mismatch_counts.append(mismatch_count)
        
        err = clopper_pearson(counts[ref_coord, read_coord], ref_total)
        errs.append(err)

    errs = np.array(errs).T  # plt.bar expects this to be 2xN

    rate_info = {'counts': mismatch_counts,
                 'rates': rates,
                 'errs': errs,
                }
    rate_info.update(compute_overall_rate(type_counts, min_q, max_q))
    
    return rate_info

def trim_paired_end_edges(mismatch_counts, trim_length):
    num_positions, _, _, _ = mismatch_counts.shape
    read_length = num_positions / 2
    min_distance_from_edge = trim_length
    far_from_edge_slice = slice(min_distance_from_edge,
                                read_length - min_distance_from_edge,
                               )
    R1_types = mismatch_counts[:read_length]
    R2_types = mismatch_counts[read_length:]
    R1_eligible_types = R1_types[far_from_edge_slice].sum(axis=0)
    R2_eligible_types = R2_types[far_from_edge_slice].sum(axis=0)
    eligible_types = R1_eligible_types + R2_eligible_types
    return eligible_types

def amplicon_mismatches(reads, expected_seq, primers, max_qual):
    per_read_counts = Counter()
    type_shape = (len(expected_seq), max_qual + 1, 6, 6)
    type_counts = np.zeros(shape=type_shape, dtype=int)

    expected_indexes = [base_to_index[b] for b in expected_seq]

    for read in reads:
        per_read_count = 0
        if len(read.seq) == len(expected_seq):
            read_indexes = [base_to_index[b] for b in read.seq]
            read_quals = fastq.decode_sanger(read.qual)
            columns = zip(read_indexes, read_quals, expected_indexes)
            for p, (read_index, q, expected_index) in enumerate(columns):
                if read_index != expected_index and q >= max_qual - 10 and p > len(primers['forward']) and p < len(expected_seq) - len(primers['reverse']):
                    per_read_count += 1
            per_read_counts[per_read_count] += 1
            if per_read_count > 4:
                continue
            for p, (read_index, q, expected_index) in enumerate(columns):
                type_counts[p, q, expected_index, read_index] += 1

    return type_counts, per_read_counts

def reference_mismatches(bam_fn,
                         genome_dir,
                         rname,
                         start=0,
                         end=None,
                         max_qual=fastq.MAX_EXPECTED_QUAL,
                         min_distance_from_edge=10,
                         need_to_split=True,
                        ):
    genome_index = genomes.get_genome_index(genome_dir)
    bam_file = pysam.Samfile(bam_fn)
    region_fetcher = genomes.build_region_fetcher(genome_dir,
                                                  load_references=True,
                                                  sam_file=bam_file,
                                                 )

    if end == None:
        end = genome_index[rname].length - 1 

    length = end - start + 1
    
    shape = (2, # forward and reverse strands
             length,
             max_qual + 1,
             len(utilities.base_order),
             len(utilities.base_order),
            )
    position_type_counts = np.zeros(shape, int)

    total_mappings = 0
    
    sequence = region_fetcher(bam_file.gettid(rname), start, end + 1).upper()

    mappings = bam_file.fetch(reference=rname, start=start, end=end)
    if need_to_split:
        split_mappings = (paired_end.split_combined_mapping(mapping, remove_soft_clipped=True)
                          for mapping in mappings
                         )
        mappings = itertools.chain.from_iterable(split_mappings)

    for mapping in mappings:
        total_mappings += 1
        if total_mappings % 10000 == 0:
            logging.info('{0:,} mappings processed'.format(total_mappings))
        #if sam.contains_soft_clipping_pysam(mapping):
        #    continue
        if mapping.seq == None:
            # Was part of an entirely-overlapping pair
            continue

        decoded_qual = fastq.decode_sanger(mapping.qual)
        
        aligned_pairs = mapping.aligned_pairs
        for pair_i, (read_i, ref_i) in enumerate(aligned_pairs):
            if read_i == None:
                # Deletion
                read_base = '-'
                # For purposes of determining how far from the edge of the read
                # the deletion is, use the read positions on either side of it.
                before_i = pair_i
                while True:
                    before_i -= 1
                    if before_i < 0:
                        break

                    previous_read_i, _ = aligned_pairs[before_i]
                    if previous_read_i != None:
                        break

                if before_i < 0:
                    logging.info(str(mapping))
                    break

                after_i = pair_i
                while True:
                    after_i += 1
                    if after_i >= len(aligned_pairs):
                        break

                    next_read_i, _ = aligned_pairs[after_i]
                    if next_read_i != None:
                        break

                if after_i >= len(aligned_pairs):
                    logging.info(str(mapping))
                    break

                from_left = previous_read_i
                from_right = mapping.qlen - 1 - next_read_i

                read_q = max_qual # Not sure if this is the right design
            else:
                from_left = read_i
                from_right = mapping.qlen - 1 - read_i
                
                read_base = mapping.seq[read_i]
                read_q = decoded_qual[read_i]
            
            from_edge = min(from_left, from_right)
            if from_edge < min_distance_from_edge:
                continue

            if ref_i == None:
                # Insertion
                ref_base = '-'

                # Assign the insertion to the ref position of the previous 
                before_i = pair_i
                while True:
                    before_i -= 1
                    if before_i < 0:
                        break

                    _, previous_ref_i = aligned_pairs[before_i]
                    if previous_ref_i != None:
                        break

                if before_i < 0:
                    logging.info(str(mapping))
                    break

                ref_i = previous_ref_i
            else:
                ref_base = region_fetcher(mapping.tid, ref_i, ref_i + 1).upper()

            if mapping.is_reverse:
                strand = 1
            else:
                strand = 0

            if not (start <= ref_i < end):
                continue

            coords = (strand,
                      ref_i - start,
                      read_q,
                      utilities.base_to_index[ref_base],
                      utilities.base_to_index[read_base],
                     )
            position_type_counts[coords] += 1

    logging.info('{0:,} total mappings processed'.format(total_mappings))

    return position_type_counts, sequence



import itertools
import log

write_file = log.write_file

def read_file(file_name):
    def process_line(line):
        description, value = line.strip().split(':')
        value = map(float, value.split(','))
        return description, value
    
    with open(file_name) as fh:
        pairs = map(process_line, fh)

    return pairs

def combine_data(first_pairs, second_pairs):
    combined_pairs = []
    zipped = itertools.izip_longest(first_pairs, second_pairs, fillvalue=(None, None))
    for (first_d, first_v), (second_d, second_v) in zipped:
        if first_d != second_d:
            raise RuntimeError, 'descriptions do not match'
        sum_v = first_v + second_v
        combined_pairs.append((first_d, sum_v))
    return combined_pairs



from collections import OrderedDict
from pathlib import Path

import Bio.SeqIO
import pandas as pd
import pysam

from . import utilities

class Record(object):
    def __init__(self, name, seq):
        self.name = name
        self.seq = seq

    def __str__(self):
        return make_record(self.name, self.seq)
    
    def reverse_complement(self):
        return Read(self.name,
                    utilities.reverse_complement(self.seq),
                   )
    
    def __getitem__(self, sl):
        return Read(self.name, self.seq[sl])
    
    def __repr__(self):
        return str(self)

    def __len__(self):
        return len(self.seq)

    def __add__(self, other):
        return Read(self.name, self.seq + other.seq)


make_record = '>{0}\n{1}\n'.format

def records(file_name):
    ''' Yields the name and sequence lines from a fasta file. '''
    file_name = Path(file_name)

    if file_name.suffix == '.ab1':
        format = 'abi'
    else:
        format = 'fasta'

    for record in Bio.SeqIO.parse(file_name, format):
        yield Record(record.name, str(record.seq).upper())

Read = Record # for backwards compatibility
reads = records

def to_dict(file_name):
    return OrderedDict((r.name, r.seq) for r in reads(file_name))

def write_dict(seq_dict, fn):
    with open(fn, 'w') as fh:
        for name, seq in seq_dict.items():
            record = Record(name, seq)
            fh.write(str(record))

    pysam.faidx(str(fn))

def load_fai(fasta_fn):
    fasta_fn = Path(fasta_fn)
    fai_fn = fasta_fn.with_suffix(fasta_fn.suffix + '.fai')

    if not fai_fn.exists():
        pysam.faidx(str(fasta_fn))

    column_names = [
        'NAME',
        'LENGTH',
        'OFFSET',
        'LINEBASES',
        'LINEWIDTH',
    ]

    fai = pd.read_csv(fai_fn, sep='\t', index_col=0, header=None, names=column_names)

    return fai



import copy

import Bio.Data.CodonTable

nucleotide_order = 'TCAG'
nucleotide_to_index = {b: i for i, b in enumerate(nucleotide_order)}
def codon_sorting_key(codon):
    return [nucleotide_to_index[b] for b in codon]

forward_table = Bio.Data.CodonTable.standard_dna_table.forward_table

non_stop_codons = sorted(forward_table, key=codon_sorting_key)
stop_codons = sorted(Bio.Data.CodonTable.standard_dna_table.stop_codons, key=codon_sorting_key)

full_forward_table = copy.copy(forward_table)
for stop_codon in stop_codons:
    full_forward_table[stop_codon] = '*'

all_codons = sorted(non_stop_codons + stop_codons, key=codon_sorting_key)
codon_to_index = {codon: i for i, codon in enumerate(all_codons)}

full_back_table = {}
for codon in all_codons:
    amino_acid = full_forward_table[codon]
    if amino_acid not in full_back_table:
        full_back_table[amino_acid] = [codon]
    else:
        full_back_table[amino_acid].append(codon)

amino_acids = sorted(full_back_table)

codon_to_amino_acid_and_index = {}
for amino_acid in full_back_table:
    for i, codon in enumerate(full_back_table[amino_acid]):
        codon_to_amino_acid_and_index[codon] = (amino_acid, i)

degeneracy = {amino_acid: len(full_back_table[amino_acid]) for amino_acid in amino_acids}

def codons_from_seq(seq):
    for i in range(0, len(seq), 3):
        yield seq[i:i + 3]

anticodon_to_codons = {
    'IGC': {'GCU', 'GCC'},
    'UGC': {'GCA', 'GCG'},
    'ICG': {'CGU', 'CGC', 'CGA'},
    'CCG': {'CGG'},
    'UCU': {'AGA'},
    'CCU': {'AGG'},
    'GUU': {'AAU', 'AAC'},
    'GUC': {'GAU', 'GAC'},
    'GCA': {'UGU', 'UGC'},
    'UUG': {'CAA'},
    'CUG': {'CAG'},
    'UUC': {'GAA'},
    'CUC': {'GAG'},
    'GCC': {'GGU', 'GGC'},
    'UCC': {'GGA'},
    'CCC': {'GGG'},
    'GUG': {'CAU', 'CAC'},
    'IAU': {'AUU', 'AUC'},
    'UAU': {'AUA'},
    'UAA': {'UUA'},
    'CAA': {'UUG'},
    'GAG': {'CUU', 'CUC'},
    'UAG': {'CUA', 'CUG'},
    'UUU': {'AAA'},
    'CUU': {'AAG'},
    'CAU': {'AUG'},
    'GAA': {'UUU', 'UUC'},
    'IGG': {'CCU', 'CCC'},
    'UGG': {'CCA', 'CCG'},
    'IGA': {'UCU', 'UCC'},
    'UGA': {'UCA'},
    'CGA': {'UCG'},
    'GCU': {'AGU', 'AGC'},
    'IGU': {'ACU', 'ACC'},
    'UGU': {'ACA'},
    'CGU': {'ACG'},
    'CCA': {'UGG'},
    'GUA': {'UAU', 'UAC'},
    'IAC': {'GUU', 'GUC'},
    'UAC': {'GUA'},
    'CAC': {'GUG'},
}

u_to_t = str.maketrans('U', 'T')
def rna_to_dna(rna):
    return rna.translate(u_to_t)

t_to_u = str.maketrans('T', 'U')
def dna_to_rna(dna):
    return dna.translate(t_to_u)

codon_to_anticodon = {}

for anticodon in anticodon_to_codons:
    dna_anticodon = rna_to_dna(anticodon)
    for codon in anticodon_to_codons[anticodon]:
        dna_codon = rna_to_dna(codon)
        codon_to_anticodon[dna_codon] = dna_anticodon



import contextlib
import datetime
import functools
import inspect
import numbers
import re
import subprocess
import sys

from collections import defaultdict
from itertools import islice, groupby, cycle, product, chain

import numpy as np
import pandas as pd
import Bio.Data.IUPACData
import scipy.optimize
import statsmodels.stats.proportion

identity = lambda x: x

mapping = Bio.Data.IUPACData.ambiguous_dna_complement
for uppercase in list(mapping):
    mapping[uppercase.lower()] = mapping[uppercase].lower()
order = ''.join(mapping)

from_bytes = order.encode()
to_bytes = ''.join(mapping[f] for f in order).encode()
complement_table = bytes.maketrans(from_bytes, to_bytes)

def complement(seq):
    return seq.translate(complement_table)

def reverse_complement(seq):
    return complement(seq)[::-1]

base_order = 'ACGTN-'
base_to_index = {b: i for i, b in enumerate(base_order)}
base_to_complement_index = {b: i for i, b in enumerate(complement(base_order))}

complement_index = {i: base_to_complement_index[b] for i, b in enumerate(base_order)}

IUPAC = {
    'A': {'A'},
    'C': {'C'},
    'G': {'G'},
    'T': {'T'},
    'M': {'A', 'C'},
    'R': {'A', 'G'},
    'W': {'A', 'T'},
    'S': {'C', 'G'},
    'Y': {'C', 'T'},
    'K': {'G', 'T'},
    'V': {'A', 'C', 'G'},
    'H': {'A', 'C', 'T'},
    'D': {'A', 'G', 'T'},
    'B': {'C', 'G', 'T'},
    'N': {'G', 'A', 'T', 'C'},
}

def counts_to_array(counts, dim=1):
    ''' Converts a dictionary with integer keys into an array. 
    Ignores negative keys.
    '''
    if dim == 1:
        if len(counts) > 0:
            biggest = max(counts)
            array = np.zeros(biggest + 1, int)
            for key, count in counts.items():
                if key >= 0:
                    array[key] = count
        else:
            array = np.array([0])
    elif dim == 2:
        if len(counts) > 0:
            larger_of_two = map(max, counts)
            biggest = max(larger_of_two)
            array = np.zeros(shape=(biggest + 1, biggest + 1), dtype=np.int)
            for key, count in counts.items():
                array[key] = count
        else:
            array = np.array([[0]])
    else:
        raise ValueError(dim)

    return array

def mean_from_histogram(histogram):
    if histogram.sum() == 0:
        return 0
    else:
        weighted_sum = np.dot(histogram, np.arange(len(histogram)))
        num_items = float(histogram.sum())
        mean = weighted_sum / num_items
    return mean

def var_from_histogram(histogram):
    if histogram.sum() == 0:
        return 0
    else:
        weighted_sum = np.dot(histogram, np.arange(len(histogram))**2)
        num_items = float(histogram.sum())
        mean_of_squared = weighted_sum / num_items
        square_of_mean = mean_from_histogram(histogram)**2
        var = mean_of_squared - square_of_mean
    return var

def mean_from_counts(counts):
    weighted_sum = sum(count * value for value, count in counts.items())
    num_items = sum(counts.values())
    return weighted_sum / num_items

def group_by(iterable, key=None, sort=False):
    ''' Groups iterable into lists of consecutive elements that are transformed
        into the same value key.
    '''
    if sort:
        iterable = sorted(iterable, key=key)

    groups = groupby(iterable, key)
    group_lists = ((value, list(iterator)) for value, iterator in groups)
    return group_lists

def round_robin(iterables):
    ''' Modified from recipe on itertools doc page credited to George Sakkis.
    '''
    pending = len(iterables)
    nexts = cycle(iter(it).next for it in iterables)
    while pending:
        try:
            for next in nexts:
                yield next()
        except StopIteration:
            pending -= 1
            # The next pending yields from nexts will be the remaining
            # non-empty iterators.
            nexts = cycle(islice(nexts, None, pending))

def line_count(file_name):
    ''' Return the number of lines in file_name by calling wc. '''
    wc_output = subprocess.check_output(['wc', '-l', file_name])
    count, _ = wc_output.strip().split()
    count = int(count)
    return count

def pairwise(s):
    ''' Returns the elements of s in overlapping pairs. '''
    return [(s[i - 1], s[i]) for i in range(1, len(s))]

def all_consecutive(s):
    ''' Returns True if the elements of s when sorted are consecutive integers. '''
    for (x, y) in pairwise(sorted(s)):
        if y - x != 1:
            return False
    return True

def empirical_cdf(values):
    ''' From stackoverflow. '''
    sorted_values = np.sort(values)
    cumulative = np.true_divide(np.arange(len(sorted_values)) + 1, len(sorted_values))
    return sorted_values, cumulative

def mers(k, include_N=False):
    if include_N:
        chars = 'TCAGN'
    else:
        chars = 'TCAG'
    return (''.join(mer) for mer in product(chars, repeat=k))

def smooth(ys, either_side):
    smoothed = pd.Series(ys).rolling(window=2 * either_side + 1, center=True, min_periods=1).mean()
    return smoothed

def reverse_dictionary(d):
    r = {v: k for k, v in d.items()}
    return r

def split_nonempty(string, delim):
    return [f for f in string.split(delim) if f]

def normalize_rows(array):
    return np.true_divide(array, array.sum(axis=1)[:, np.newaxis])

def reverse_enumerate(list_like):
    reversed_range = range(len(list_like) - 1, -1, -1)
    return zip(reversed_range, list_like[::-1])

def memoized_property(f):
    @property
    @functools.wraps(f)
    def memoized_f(self):
        attr_name = f'_memoized_{f.__name__}'
        
        if not hasattr(self, attr_name):
            setattr(self, attr_name, f(self))
        
        return getattr(self, attr_name)
    
    return memoized_f

def memoized_with_args(f):
    @functools.wraps(f)
    def memoized_f(self, *args):
        attr_name = f'_memoized_{f.__name__}'
        if not hasattr(self, attr_name):
            setattr(self, attr_name, {})

        already_computed = getattr(self, attr_name)
        if args in already_computed:
            value = already_computed[args]
        else:
            value = f(self, *args)
            already_computed[args] = value

        return value

    return memoized_f

def memoized_with_kwargs(f):
    @functools.wraps(f)
    def memoized_f(self, **kwargs):
        argspec = inspect.getfullargspec(f)
        if argspec.defaults is not None or len(argspec.args) != 1:
            raise ValueError('all arguments must be keyword-only')

        if argspec.args[0] != 'self':
            raise ValueError('first arg name must be self')

        if argspec.kwonlydefaults is not None:
            for key, value in argspec.kwonlydefaults.items():
                kwargs.setdefault(key, value)

        all_args = tuple([(key, value) for key, value in sorted(kwargs.items())])

        attr_name = f'_memoized_{f.__name__}'
        if not hasattr(self, attr_name):
            setattr(self, attr_name, {})

        already_computed = getattr(self, attr_name)
        if all_args in already_computed:
            value = already_computed[all_args]
        else:
            value = f(self, **kwargs)
            already_computed[all_args] = value

        return value

    return memoized_f

def reservoir_sample(iterable, n):
    sampler = ReservoirSampler(n)
    for item in iterable:
        sampler.add(item)
    return sampler.sample

class ReservoirSampler():
    def __init__(self, sample_size):
        self.sample = []
        self.sample_size = sample_size
        self.num_items_seen = 0

    def add(self, item):
        if self.num_items_seen < self.sample_size:
            self.sample.append(item)
        else:
            j = np.random.randint(self.num_items_seen + 1)
            if j < self.sample_size:
                self.sample[j] = item

        self.num_items_seen += 1

def chunks(iterable, n):
    '''from https://stackoverflow.com/a/29524877
    Note: this only works if you actually consume each chunk before getting the
    next one.
    '''
    iterable = iter(iterable)
    while True:
        try:
            first = next(iterable)
        except StopIteration:
            return

        rest = islice(iterable, n - 1)

        yield chain([first], rest)

def list_chunks(full_list, n):
    starts = np.arange(0, len(full_list), n)
    ends = starts + n
    chunks = [full_list[start:end] for start, end in zip(starts, ends)]
    return chunks

@contextlib.contextmanager
def possibly_fn(fn=None):
    # from https://stackoverflow.com/a/22264583
    if fn is not None:
        writer = open(str(fn), 'w')
    else:
        writer = sys.stdout

    yield writer

    if fn != None: writer.close()

def clopper_pearson(x, n, alpha=0.05, return_format='delta'):
    if n == 0:
        return 0., 0.

    elif x == 0:
        cdf = lambda t: scipy.stats.binom.cdf(x, n, t) - alpha
        u = scipy.optimize.brentq(cdf, 0, 1)
        l = 0.
        mle = 0.

    elif x == n:
        sf = lambda t: scipy.stats.binom.sf(x - 1, n, t) - alpha
        u = 1.
        l = scipy.optimize.brentq(sf, 0, 1)
        mle = 1.

    else:
        cdf = lambda t: scipy.stats.binom.cdf(x, n, t) - (alpha / 2)
        sf = lambda t: scipy.stats.binom.sf(x - 1, n, t) - (alpha / 2)
        u = scipy.optimize.brentq(cdf, 0, 1)
        l = scipy.optimize.brentq(sf, 0, 1)
        mle = float(x) / n

    if return_format == 'delta':
        return mle - l, u - mle
    elif return_format == 'absolute':
        return l, u
    else:
        raise ValueError(return_format)

def clopper_pearson_fast(xs, ns, alpha=0.05):
    if isinstance(xs, numbers.Number):
        alphas = alpha
    else:
        alphas = np.full(len(xs), alpha)
    #alphas[(xs == 0) | (xs == ns)] = 2 * alpha
    lowers, uppers = statsmodels.stats.proportion.proportion_confint(xs, ns, alpha=alphas, method='beta')

    return lowers, uppers

def homopolymer_lengths(seq, b):
    locations = []
    
    i = 0
    while True:
        # Advance until you find a b
        while  i < len(seq) and seq[i] != b:
            i += 1
        # If you never did, you are done.
        if i == len(seq):
            break
            
        start = i
        # Advance until the poly-b stretch starting at i ends.
        while i < len(seq) and seq[i] == b:
            i += 1
        
        length = i - start
        locations.append((start, length))
        
        if i == len(seq):
            break
    
    return locations

def get_one_mismatch_resolver(sample_indices, prioritize_exact_matches=False):
    ''' If prioritize_exact_matches is True, in cases where two sequences are 
    hamming distance one from each other, allow the exact sequence for each to
    uniquely resolve to its label (i.e. don't register a mismatched seq
    if it is also an exactly matching seq).
    '''
    def get_all_one_mismatch(seq):
        seqs = set()
        for i in range(len(seq)):
            for b in 'TCAGN':
                seqs.add(seq[:i] + b + seq[i + 1:])

        return seqs

    all_exact_seqs = set()
    for name, seqs in sample_indices.items():
        if not isinstance(seqs, list):
            seqs = [seqs]
        all_exact_seqs.update(seqs)
    
    resolver = defaultdict(list)
    for name, seqs in sample_indices.items():
        if not isinstance(seqs, list):
            seqs = [seqs]
        for seq in seqs:
            for one_mismatch_seq in get_all_one_mismatch(seq):
                if prioritize_exact_matches and one_mismatch_seq != seq and one_mismatch_seq in all_exact_seqs:
                    continue
                resolver[one_mismatch_seq].append(name)
            
    resolver = {seq: set(names) for seq, names in resolver.items()}

    return resolver

def current_time_string():
    return f'{datetime.datetime.now():%Y-%m-%d %H:%M:%S}'

def find_all_substring_starts(target_sequence, substring):
    # lookahead necessary in the event of partially-overlapping matches
    return [match.start() for match in re.finditer(f'(?={substring})', target_sequence)]

def partial_class(cls, *args, **kwargs):
    ''' from https://stackoverflow.com/a/38911383 '''

    class PartialClass(cls):
        __init__ = functools.partialmethod(cls.__init__, *args, **kwargs)

    return PartialClass


import ipywidgets
import IPython.display

# from http://chris-said.io/
toggle = '''
<script>
  function code_toggle() {
    if (code_shown){
      $('div.input').hide('100');
      $('#toggleButton').val('Show Code')
    } else {
      $('div.input').show('100');
      $('#toggleButton').val('Hide Code')
    }
    code_shown = !code_shown
  }

  $( document ).ready(function(){
    code_shown=false;
    $('div.input').hide()
  });
</script>
<form action="javascript:code_toggle()"><input type="submit" id="toggleButton" value="Show Code"></form>
'''

toggle_cell = IPython.display.HTML(toggle)

def binary_widget(default=True):
    values = [int(default), int(not default)]
    labels = [str(default), str(not default)]
    return ipywidgets.RadioWidget(values, labels=labels)



import sys
import os
from hits import utilities

lonestar_template = '''\
#!/bin/bash

#$ -N {job_name}
#$ -pe 12way {num_cores}
#$ -q normal
#$ -o {job_name}.o$JOB_ID
#$ -l h_rt={time}
#$ -V
#$ -M jah@ices.utexas.edu
#$ -m be
{hold_line}
#$ -cwd
#$ -A SeqMap1
'''

stampede_template = '''\
#!/bin/bash

#SBATCH -J {job_name}
#SBATCH -n {num_cores}
#SBATCH -p normal
#SBATCH -o {job_name}.o%j
#SBATCH -e {job_name}.e%j
#SBATCH -t {time}
{hold_line}
#SBATCH -A SeqMap1
#SBATCH --mail-type=ALL
#SBATCH --mail-user=jah@ices.utexas.edu
'''

common_template = '''\
module load launcher
module load samtools
module load bwa

export EXECUTABLE=$TACC_LAUNCHER_DIR/init_launcher 
export CONTROL_FILE={commands_file_name}
export WORKDIR=.
 
# Variable description:
#
#  EXECUTABLE     = full path to the job launcher executable
#  CONTROL_FILE   = text input file which specifies
#                   executable for each process
#                   (should be located in WORKDIR)
#  WORKDIR        = location of working directory

if [ ! -e $WORKDIR ]; then
    echo " "
    echo "Error: unable to change to working directory."
	echo "       $WORKDIR"
	echo " "
	echo "Job not submitted."
	exit
fi

if [ ! -f $EXECUTABLE ]; then
	echo " "
	echo "Error: unable to find launcher executable $EXECUTABLE."
	echo " "
	echo "Job not submitted."
	exit
fi

if [ ! -f $WORKDIR/$CONTROL_FILE ]; then
	echo " "
	echo "Error: unable to find input control file $CONTROL_FILE."
	echo " "
	echo "Job not submitted."
	exit
fi

cd $WORKDIR/
echo " WORKING DIR:   $WORKDIR/"

$TACC_LAUNCHER_DIR/paramrun $EXECUTABLE $CONTROL_FILE

echo " "
echo " Parameteric Job Complete"

{optional_finish}
python $HOME/projects/mutations/code/Parallel/tacc_email_error.py {job_name} {job_dir} {job_id_var}
'''

def create(job_name,
           full_commands_file_name,
           time='1:00:00',
           optional_finish='# No optional finish',
           hold_jid=None,
          ):
    job_dir, commands_file_name = os.path.split(full_commands_file_name)

    hostname = os.environ['HOSTNAME']
    if 'ls4' in hostname:
        cores_per_node = 12
        launcher_file_name = '{0}/launcher_{1}.sge'.format(job_dir, job_name)
        specific_template = lonestar_template
        hold_template = '#$ -hold_jid {0}'
        job_id_var = '$JOB_ID'
    elif 'stampede' in hostname:
        cores_per_node = 16
        launcher_file_name = '{0}/launcher_{1}.slurm'.format(job_dir, job_name)
        specific_template = stampede_template
        hold_template = '#SBATCH -d afterok:{0}'
        job_id_var = '$SLURM_JOB_ID'
    
    num_cores = utilities.line_count(full_commands_file_name)
    while num_cores % cores_per_node != 0:
        num_cores += 1

    if hold_jid:
        hold_line = hold_template.format(hold_jid)
    else:
        hold_line = '# Not holding'
    
    specific_part = specific_template.format(job_name=job_name,
                                             num_cores=num_cores,
                                             time=time,
                                             hold_line=hold_line,
                                            )
    common_part = common_template.format(job_name=job_name,
                                         job_dir=job_dir,
                                         commands_file_name=commands_file_name,
                                         optional_finish=optional_finish,
                                         job_id_var=job_id_var,
                                        )
    
    with open(launcher_file_name, 'w') as launcher_file:
        launcher_file.write(specific_part + common_part)

    return launcher_file_name
